// -------------------------------------------------------------------------------
/*
   m4dMetricBarriolaVilenkin.cpp
 
  Copyright (c) 2009  Thomas Mueller, Frank Grave
 
 
   This file is part of the m4d-library.
 
   The m4d-library is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
 
   The m4d-library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with the m4d-library.  If not, see <http://www.gnu.org/licenses/>.
 
*/
// -------------------------------------------------------------------------------

#include "m4dMetricBarriolaVilenkin.h"

namespace m4d
{

/*! Standard constructor for the BarriolaVilenkin metric.
 *
 * \param  k : monopol parameter.
 */
MetricBarriolaVilenkin :: MetricBarriolaVilenkin( double k )
{
  mMetricName = "BarriolaVilenkin";
  setCoordType(enum_coordinate_spherical);
  
  mPhysicalUnits = enum_physical_constants_geom;
  mSpeedOfLight = 1.0;
  mGravConstant = 1.0;

  addParam("k",k); 
  mk = k;
  
  mLocTeds.push_back(enum_nat_tetrad_static);  

  mDrawTypes.push_back(enum_draw_embedding);
  mDrawTypes.push_back(enum_draw_effpoti);

  /*  parameters for the embedding diagram  */
  if (!mEmbParam.empty())
    mEmbParam.clear();
  mHaveEmbedding = true;

  mEmb_rmin    = 0.0;
  mEmb_rmax    = 10.0;
  mEmb_r_num   = 20.0;
  mEmb_phi_num = 40.0;
  mEmb_rstep = (mEmb_rmax-mEmb_rmin)/mEmb_r_num;
  mEmb_phistep = 2.0*M_PI/mEmb_phi_num;
  addEmbeddingParam("emb_rmin",mEmb_rmin);
  addEmbeddingParam("emb_rmax",mEmb_rmax);
  addEmbeddingParam("emb_r_num",20.0);
  addEmbeddingParam("emb_phi_num",40.0);

  setStandardValues();
}

MetricBarriolaVilenkin :: ~MetricBarriolaVilenkin()
{
}


// *********************************** public methods ******************************
/*!  Calculate the contravariant metric components at position 'pos'.
 *
 *  \param pos : pointer to position.
 */
bool 
MetricBarriolaVilenkin :: calculateMetric ( const double* pos )
{
  double r = pos[1];
  double theta = pos[2];
  double c = mSpeedOfLight;

  double t1 = c*c;
  double t2 = mk*mk;
  double t3 = r*r;
  double t4 = t2*t3;
  double t5 = sin(theta);
  double t6 = t5*t5;

      g_compts[0][0] = -t1;
      g_compts[0][1] = 0.0;
      g_compts[0][2] = 0.0;
      g_compts[0][3] = 0.0;
      g_compts[1][0] = 0.0;
      g_compts[1][1] = 1.0;
      g_compts[1][2] = 0.0;
      g_compts[1][3] = 0.0;
      g_compts[2][0] = 0.0;
      g_compts[2][1] = 0.0;
      g_compts[2][2] = t4;
      g_compts[2][3] = 0.0;
      g_compts[3][0] = 0.0;
      g_compts[3][1] = 0.0;
      g_compts[3][2] = 0.0;
      g_compts[3][3] = t4*t6;

  return true;
}

/*! Calculate the Christoffel symbols of the second kind at position 'pos'.
 *
 *  \param pos : pointer to position.
 */
bool 
MetricBarriolaVilenkin :: calculateChristoffels  ( const double* pos )
{
  double r = pos[1];
  double theta = pos[2];

  double t1 = 1/r;
  double t2 = mk*mk;
  double t3 = t2*r;
  double t4 = sin(theta);
  double t6 = cos(theta);
  double t7 = 1/t4*t6;
  double t8 = t4*t4;

      christoffel[0][0][0] = 0.0;
      christoffel[0][0][1] = 0.0;
      christoffel[0][0][2] = 0.0;
      christoffel[0][0][3] = 0.0;
      christoffel[0][1][0] = 0.0;
      christoffel[0][1][1] = 0.0;
      christoffel[0][1][2] = 0.0;
      christoffel[0][1][3] = 0.0;
      christoffel[0][2][0] = 0.0;
      christoffel[0][2][1] = 0.0;
      christoffel[0][2][2] = 0.0;
      christoffel[0][2][3] = 0.0;
      christoffel[0][3][0] = 0.0;
      christoffel[0][3][1] = 0.0;
      christoffel[0][3][2] = 0.0;
      christoffel[0][3][3] = 0.0;
      christoffel[1][0][0] = 0.0;
      christoffel[1][0][1] = 0.0;
      christoffel[1][0][2] = 0.0;
      christoffel[1][0][3] = 0.0;
      christoffel[1][1][0] = 0.0;
      christoffel[1][1][1] = 0.0;
      christoffel[1][1][2] = 0.0;
      christoffel[1][1][3] = 0.0;
      christoffel[1][2][0] = 0.0;
      christoffel[1][2][1] = 0.0;
      christoffel[1][2][2] = t1;
      christoffel[1][2][3] = 0.0;
      christoffel[1][3][0] = 0.0;
      christoffel[1][3][1] = 0.0;
      christoffel[1][3][2] = 0.0;
      christoffel[1][3][3] = t1;
      christoffel[2][0][0] = 0.0;
      christoffel[2][0][1] = 0.0;
      christoffel[2][0][2] = 0.0;
      christoffel[2][0][3] = 0.0;
      christoffel[2][1][0] = 0.0;
      christoffel[2][1][1] = 0.0;
      christoffel[2][1][2] = t1;
      christoffel[2][1][3] = 0.0;
      christoffel[2][2][0] = 0.0;
      christoffel[2][2][1] = -t3;
      christoffel[2][2][2] = 0.0;
      christoffel[2][2][3] = 0.0;
      christoffel[2][3][0] = 0.0;
      christoffel[2][3][1] = 0.0;
      christoffel[2][3][2] = 0.0;
      christoffel[2][3][3] = t7;
      christoffel[3][0][0] = 0.0;
      christoffel[3][0][1] = 0.0;
      christoffel[3][0][2] = 0.0;
      christoffel[3][0][3] = 0.0;
      christoffel[3][1][0] = 0.0;
      christoffel[3][1][1] = 0.0;
      christoffel[3][1][2] = 0.0;
      christoffel[3][1][3] = t1;
      christoffel[3][2][0] = 0.0;
      christoffel[3][2][1] = 0.0;
      christoffel[3][2][2] = 0.0;
      christoffel[3][2][3] = t7;
      christoffel[3][3][0] = 0.0;
      christoffel[3][3][1] = -t3*t8;
      christoffel[3][3][2] = -t4*t6;
      christoffel[3][3][3] = 0.0; 

  return true;
}


/*! Calculate Jacobi matrix. 
 *
 *  \param pos : pointer to position.
 */
bool 
MetricBarriolaVilenkin :: calculateChrisD ( const double* pos )
{
  double r = pos[1];
  double theta = pos[2];

  double t1 = r*r;
  double t2 = 1/t1;
  double t3 = mk*mk;
  double t4 = cos(theta);
  double t5 = t4*t4;
  double t6 = sin(theta);
  double t7 = t6*t6;
  double t10 = (t5+t7)/t7;

      chrisD[0][0][0][0] = 0.0;
      chrisD[0][0][0][1] = 0.0;
      chrisD[0][0][0][2] = 0.0;
      chrisD[0][0][0][3] = 0.0;
      chrisD[0][0][1][0] = 0.0;
      chrisD[0][0][1][1] = 0.0;
      chrisD[0][0][1][2] = 0.0;
      chrisD[0][0][1][3] = 0.0;
      chrisD[0][0][2][0] = 0.0;
      chrisD[0][0][2][1] = 0.0;
      chrisD[0][0][2][2] = 0.0;
      chrisD[0][0][2][3] = 0.0;
      chrisD[0][0][3][0] = 0.0;
      chrisD[0][0][3][1] = 0.0;
      chrisD[0][0][3][2] = 0.0;
      chrisD[0][0][3][3] = 0.0;
      chrisD[0][1][0][0] = 0.0;
      chrisD[0][1][0][1] = 0.0;
      chrisD[0][1][0][2] = 0.0;
      chrisD[0][1][0][3] = 0.0;
      chrisD[0][1][1][0] = 0.0;
      chrisD[0][1][1][1] = 0.0;
      chrisD[0][1][1][2] = 0.0;
      chrisD[0][1][1][3] = 0.0;
      chrisD[0][1][2][0] = 0.0;
      chrisD[0][1][2][1] = 0.0;
      chrisD[0][1][2][2] = 0.0;
      chrisD[0][1][2][3] = 0.0;
      chrisD[0][1][3][0] = 0.0;
      chrisD[0][1][3][1] = 0.0;
      chrisD[0][1][3][2] = 0.0;
      chrisD[0][1][3][3] = 0.0;
      chrisD[0][2][0][0] = 0.0;
      chrisD[0][2][0][1] = 0.0;
      chrisD[0][2][0][2] = 0.0;
      chrisD[0][2][0][3] = 0.0;
      chrisD[0][2][1][0] = 0.0;
      chrisD[0][2][1][1] = 0.0;
      chrisD[0][2][1][2] = 0.0;
      chrisD[0][2][1][3] = 0.0;
      chrisD[0][2][2][0] = 0.0;
      chrisD[0][2][2][1] = 0.0;
      chrisD[0][2][2][2] = 0.0;
      chrisD[0][2][2][3] = 0.0;
      chrisD[0][2][3][0] = 0.0;
      chrisD[0][2][3][1] = 0.0;
      chrisD[0][2][3][2] = 0.0;
      chrisD[0][2][3][3] = 0.0;
      chrisD[0][3][0][0] = 0.0;
      chrisD[0][3][0][1] = 0.0;
      chrisD[0][3][0][2] = 0.0;
      chrisD[0][3][0][3] = 0.0;
      chrisD[0][3][1][0] = 0.0;
      chrisD[0][3][1][1] = 0.0;
      chrisD[0][3][1][2] = 0.0;
      chrisD[0][3][1][3] = 0.0;
      chrisD[0][3][2][0] = 0.0;
      chrisD[0][3][2][1] = 0.0;
      chrisD[0][3][2][2] = 0.0;
      chrisD[0][3][2][3] = 0.0;
      chrisD[0][3][3][0] = 0.0;
      chrisD[0][3][3][1] = 0.0;
      chrisD[0][3][3][2] = 0.0;
      chrisD[0][3][3][3] = 0.0;
      chrisD[1][0][0][0] = 0.0;
      chrisD[1][0][0][1] = 0.0;
      chrisD[1][0][0][2] = 0.0;
      chrisD[1][0][0][3] = 0.0;
      chrisD[1][0][1][0] = 0.0;
      chrisD[1][0][1][1] = 0.0;
      chrisD[1][0][1][2] = 0.0;
      chrisD[1][0][1][3] = 0.0;
      chrisD[1][0][2][0] = 0.0;
      chrisD[1][0][2][1] = 0.0;
      chrisD[1][0][2][2] = 0.0;
      chrisD[1][0][2][3] = 0.0;
      chrisD[1][0][3][0] = 0.0;
      chrisD[1][0][3][1] = 0.0;
      chrisD[1][0][3][2] = 0.0;
      chrisD[1][0][3][3] = 0.0;
      chrisD[1][1][0][0] = 0.0;
      chrisD[1][1][0][1] = 0.0;
      chrisD[1][1][0][2] = 0.0;
      chrisD[1][1][0][3] = 0.0;
      chrisD[1][1][1][0] = 0.0;
      chrisD[1][1][1][1] = 0.0;
      chrisD[1][1][1][2] = 0.0;
      chrisD[1][1][1][3] = 0.0;
      chrisD[1][1][2][0] = 0.0;
      chrisD[1][1][2][1] = 0.0;
      chrisD[1][1][2][2] = 0.0;
      chrisD[1][1][2][3] = 0.0;
      chrisD[1][1][3][0] = 0.0;
      chrisD[1][1][3][1] = 0.0;
      chrisD[1][1][3][2] = 0.0;
      chrisD[1][1][3][3] = 0.0;
      chrisD[1][2][0][0] = 0.0;
      chrisD[1][2][0][1] = 0.0;
      chrisD[1][2][0][2] = 0.0;
      chrisD[1][2][0][3] = 0.0;
      chrisD[1][2][1][0] = 0.0;
      chrisD[1][2][1][1] = 0.0;
      chrisD[1][2][1][2] = 0.0;
      chrisD[1][2][1][3] = 0.0;
      chrisD[1][2][2][0] = 0.0;
      chrisD[1][2][2][1] = -t2;
      chrisD[1][2][2][2] = 0.0;
      chrisD[1][2][2][3] = 0.0;
      chrisD[1][2][3][0] = 0.0;
      chrisD[1][2][3][1] = 0.0;
      chrisD[1][2][3][2] = 0.0;
      chrisD[1][2][3][3] = 0.0;
      chrisD[1][3][0][0] = 0.0;
      chrisD[1][3][0][1] = 0.0;
      chrisD[1][3][0][2] = 0.0;
      chrisD[1][3][0][3] = 0.0;
      chrisD[1][3][1][0] = 0.0;
      chrisD[1][3][1][1] = 0.0;
      chrisD[1][3][1][2] = 0.0;
      chrisD[1][3][1][3] = 0.0;
      chrisD[1][3][2][0] = 0.0;
      chrisD[1][3][2][1] = 0.0;
      chrisD[1][3][2][2] = 0.0;
      chrisD[1][3][2][3] = 0.0;
      chrisD[1][3][3][0] = 0.0;
      chrisD[1][3][3][1] = -t2;
      chrisD[1][3][3][2] = 0.0;
      chrisD[1][3][3][3] = 0.0;
      chrisD[2][0][0][0] = 0.0;
      chrisD[2][0][0][1] = 0.0;
      chrisD[2][0][0][2] = 0.0;
      chrisD[2][0][0][3] = 0.0;
      chrisD[2][0][1][0] = 0.0;
      chrisD[2][0][1][1] = 0.0;
      chrisD[2][0][1][2] = 0.0;
      chrisD[2][0][1][3] = 0.0;
      chrisD[2][0][2][0] = 0.0;
      chrisD[2][0][2][1] = 0.0;
      chrisD[2][0][2][2] = 0.0;
      chrisD[2][0][2][3] = 0.0;
      chrisD[2][0][3][0] = 0.0;
      chrisD[2][0][3][1] = 0.0;
      chrisD[2][0][3][2] = 0.0;
      chrisD[2][0][3][3] = 0.0;
      chrisD[2][1][0][0] = 0.0;
      chrisD[2][1][0][1] = 0.0;
      chrisD[2][1][0][2] = 0.0;
      chrisD[2][1][0][3] = 0.0;
      chrisD[2][1][1][0] = 0.0;
      chrisD[2][1][1][1] = 0.0;
      chrisD[2][1][1][2] = 0.0;
      chrisD[2][1][1][3] = 0.0;
      chrisD[2][1][2][0] = 0.0;
      chrisD[2][1][2][1] = -t2;
      chrisD[2][1][2][2] = 0.0;
      chrisD[2][1][2][3] = 0.0;
      chrisD[2][1][3][0] = 0.0;
      chrisD[2][1][3][1] = 0.0;
      chrisD[2][1][3][2] = 0.0;
      chrisD[2][1][3][3] = 0.0;
      chrisD[2][2][0][0] = 0.0;
      chrisD[2][2][0][1] = 0.0;
      chrisD[2][2][0][2] = 0.0;
      chrisD[2][2][0][3] = 0.0;
      chrisD[2][2][1][0] = 0.0;
      chrisD[2][2][1][1] = -t3;
      chrisD[2][2][1][2] = 0.0;
      chrisD[2][2][1][3] = 0.0;
      chrisD[2][2][2][0] = 0.0;
      chrisD[2][2][2][1] = 0.0;
      chrisD[2][2][2][2] = 0.0;
      chrisD[2][2][2][3] = 0.0;
      chrisD[2][2][3][0] = 0.0;
      chrisD[2][2][3][1] = 0.0;
      chrisD[2][2][3][2] = 0.0;
      chrisD[2][2][3][3] = 0.0;
      chrisD[2][3][0][0] = 0.0;
      chrisD[2][3][0][1] = 0.0;
      chrisD[2][3][0][2] = 0.0;
      chrisD[2][3][0][3] = 0.0;
      chrisD[2][3][1][0] = 0.0;
      chrisD[2][3][1][1] = 0.0;
      chrisD[2][3][1][2] = 0.0;
      chrisD[2][3][1][3] = 0.0;
      chrisD[2][3][2][0] = 0.0;
      chrisD[2][3][2][1] = 0.0;
      chrisD[2][3][2][2] = 0.0;
      chrisD[2][3][2][3] = 0.0;
      chrisD[2][3][3][0] = 0.0;
      chrisD[2][3][3][1] = 0.0;
      chrisD[2][3][3][2] = -t10;
      chrisD[2][3][3][3] = 0.0;
      chrisD[3][0][0][0] = 0.0;
      chrisD[3][0][0][1] = 0.0;
      chrisD[3][0][0][2] = 0.0;
      chrisD[3][0][0][3] = 0.0;
      chrisD[3][0][1][0] = 0.0;
      chrisD[3][0][1][1] = 0.0;
      chrisD[3][0][1][2] = 0.0;
      chrisD[3][0][1][3] = 0.0;
      chrisD[3][0][2][0] = 0.0;
      chrisD[3][0][2][1] = 0.0;
      chrisD[3][0][2][2] = 0.0;
      chrisD[3][0][2][3] = 0.0;
      chrisD[3][0][3][0] = 0.0;
      chrisD[3][0][3][1] = 0.0;
      chrisD[3][0][3][2] = 0.0;
      chrisD[3][0][3][3] = 0.0;
      chrisD[3][1][0][0] = 0.0;
      chrisD[3][1][0][1] = 0.0;
      chrisD[3][1][0][2] = 0.0;
      chrisD[3][1][0][3] = 0.0;
      chrisD[3][1][1][0] = 0.0;
      chrisD[3][1][1][1] = 0.0;
      chrisD[3][1][1][2] = 0.0;
      chrisD[3][1][1][3] = 0.0;
      chrisD[3][1][2][0] = 0.0;
      chrisD[3][1][2][1] = 0.0;
      chrisD[3][1][2][2] = 0.0;
      chrisD[3][1][2][3] = 0.0;
      chrisD[3][1][3][0] = 0.0;
      chrisD[3][1][3][1] = -t2;
      chrisD[3][1][3][2] = 0.0;
      chrisD[3][1][3][3] = 0.0;
      chrisD[3][2][0][0] = 0.0;
      chrisD[3][2][0][1] = 0.0;
      chrisD[3][2][0][2] = 0.0;
      chrisD[3][2][0][3] = 0.0;
      chrisD[3][2][1][0] = 0.0;
      chrisD[3][2][1][1] = 0.0;
      chrisD[3][2][1][2] = 0.0;
      chrisD[3][2][1][3] = 0.0;
      chrisD[3][2][2][0] = 0.0;
      chrisD[3][2][2][1] = 0.0;
      chrisD[3][2][2][2] = 0.0;
      chrisD[3][2][2][3] = 0.0;
      chrisD[3][2][3][0] = 0.0;
      chrisD[3][2][3][1] = 0.0;
      chrisD[3][2][3][2] = -t10;
      chrisD[3][2][3][3] = 0.0;
      chrisD[3][3][0][0] = 0.0;
      chrisD[3][3][0][1] = 0.0;
      chrisD[3][3][0][2] = 0.0;
      chrisD[3][3][0][3] = 0.0;
      chrisD[3][3][1][0] = 0.0;
      chrisD[3][3][1][1] = -t3*t7;
      chrisD[3][3][1][2] = -2.0*t3*r*t6*t4;
      chrisD[3][3][1][3] = 0.0;
      chrisD[3][3][2][0] = 0.0;
      chrisD[3][3][2][1] = 0.0;
      chrisD[3][3][2][2] = -t5+t7;
      chrisD[3][3][2][3] = 0.0;
      chrisD[3][3][3][0] = 0.0;
      chrisD[3][3][3][1] = 0.0;
      chrisD[3][3][3][2] = 0.0;
      chrisD[3][3][3][3] = 0.0;
    
    return true;
}

/*! Transform local 4-direction to coordinate 4-direction.
 *
 *  \param  pos  :  pointer to position array.
 *  \param  ldir :  pointer to local direction array.
 *  \param  dir  :  pointer to calculated coordinate direction array.
 *  \param  type :  type of tetrad.
 */
void
MetricBarriolaVilenkin :: localToCoord ( const double* pos, const double* ldir, double* dir, 
                                         enum_nat_tetrad_type   )
{
  double r     = pos[1];
  double theta = pos[2];
  
  dir[0] = ldir[0]/mSpeedOfLight;
  dir[1] = ldir[1];
  dir[2] = ldir[2]/(mk*r);
  dir[3] = ldir[3]/(mk*r*sin(theta));
}

/*! Transform coordinate 4-direction to local 4-direction.
 *
 *  \param  pos  :  pointer to position array.
 *  \param  cdir :  pointer to coordinate direction.
 *  \param  ldir :  pointer to calculated local direction array.
 *  \param  type :  type of tetrad.
 */
void
MetricBarriolaVilenkin :: coordToLocal ( const double* pos, const double* cdir, double* ldir, 
                                         enum_nat_tetrad_type   )
{
  double r     = pos[1];
  double theta = pos[2];
    
  ldir[0] = cdir[0]*mSpeedOfLight;
  ldir[1] = cdir[1];
  ldir[2] = cdir[2]*mk*r;
  ldir[3] = cdir[3]*mk*r*sin(theta);
}

/*! Test break condition.
 *
 *  \param pos    : pointer to position array.
 *  \return false : position is valid.
 */
bool 
MetricBarriolaVilenkin :: breakCondition ( const double* )
{
  return false;
}

/*! Calculate right hand side of the geodesic equation in first order form.
 *
 *  \param  y[]   : pointer to position and direction coordinates.
 *  \param  dydx[] : pointer to right side of geodesic equation.
 */
bool 
MetricBarriolaVilenkin :: calcDerivs ( const double y[], double dydx[] )
{
  dydx[0] = y[4];
  dydx[1] = y[5];
  dydx[2] = y[6];
  dydx[3] = y[7];

  double r = y[1];
  double theta = y[2];
  double st = sin(theta);
  double ct = cos(theta);

  dydx[4] = 0.0;
  dydx[5] = mk*mk*r*(y[6]*y[6] + st*st*y[7]*y[7]);
  dydx[6] = -2.0/r*y[5]*y[6] + st*ct*y[7]*y[7];
  dydx[7] = -2.0/r*y[5]*y[7] - 2.0*ct/st*y[6]*y[7];
  
  return true;
}


/*! Tests whether the constraint equation is fulfilled.
 *
 *  The constraint equation for lightlike and timelike geodesics reads:  
 \verbatim
     sum = g_{\mu\nu} dot(x)^{\mu} dot(x)^{\nu} - kappa c^2 = 0.
 \endverbatim
 *  \param  y[]   : pointer to position and direction coordinates.
 *  \param  kappa : timelike (-1.0), lightlike (0.0).
 *  \return double : sum.
 */
double 
MetricBarriolaVilenkin :: testConstraint ( const double y[], const double kappa )
{
  double r     = y[1];
  double theta = y[2];
  double cm = 1.0/mSpeedOfLight;
  
  // Scale the directions with the speed of light before doubling them !!
  double dt  = y[4];
  double dr  = y[5]*cm;
  double dth = y[6]*cm;
  double dph = y[7]*cm;

  double sum = -kappa;
  sum += -dt*dt + dr*dr + mk*mk*r*r*(dth*dth + sin(theta)*sin(theta)*dph*dph);
  return sum;
}


/*! Set parameter 'pName' to 'val'.
 *
 *  Set 'mass' parameter and adjust Schwarzschild radius  rs=2GM/c^2.
 */
bool 
MetricBarriolaVilenkin :: setParam ( std::string pName, double val )
{
  if (Metric::setParam(pName,val))
    mk = val;
  
  return true;
}


/*! Transform point p to embedding coordinates.
 *
 *  \param p  : point to be transformed.
 *  \param ep : reference to 'embedded' point.
 *  \return true : success.
 *  \return false : otherwise.
 */
bool
MetricBarriolaVilenkin :: transToEmbedding ( vec4 p, vec4 &ep )
{
  vec4 cp;
  transToPseudoCart(p,cp);

  double r = p[1];
  double x = cp[1];
  double y = cp[2];
  double z;

  if (fabs(mk)<=1.0)
  {
    z = sqrt(1.0-mk*mk)*r;
    ep = vec4(p[0],x,y,z);
    return true;
  }
  return false;
}

/*! Set embedding parameters.
 *
 *  \param  name : embedding parameter name.
 *  \param  val  : embedding parameter value.
 *  \return true  : success.
 *  \return false : parameter not valid.
 */
bool
MetricBarriolaVilenkin :: setEmbeddingParam    ( std::string name, double val )
{
  Metric::setEmbeddingParam(name,val);

  if (name=="emb_rmin")
  {
    mEmb_rmin = val;
  }
  else if (name=="emb_rmax")
  {
    mEmb_rmax = val;
  }
  else if (name=="emb_r_num")
  {
    mEmb_r_num = val;
  }
  else if (name=="emb_phi_num")
  {
    mEmb_phi_num = val;
  }
  return testEmbeddingParams();
}

/*! Test embedding parameters
 *  \return  true : all parameters are ok
 *  \return  false : at least one parameter had to be adjusted.
 */
bool
MetricBarriolaVilenkin :: testEmbeddingParams  ( )
{
  bool allOk = true;
  if (mEmb_rmin<0.0)
  {
    mEmb_rmin = 0.0;
    allOk &= false;
  }
  if (mEmb_rmax<0.0)
  {
    mEmb_rmax = 0.0;
    allOk &= false;
  }
  if (mEmb_r_num<2.0)
  {
    mEmb_r_num = 2.0;
    allOk &= false;
  }

  if (mEmb_phi_num<4)
  {
    mEmb_phi_num = 4;
    allOk &= false;
  }
  return allOk;
}

/*! Generate vertices for the embedding diagram.
 *
 *  \param verts : reference to vector of vertices.
 *  \param indices : reference to vector of indices.
 *  \param numElems : number of elements in a strip.
 *  \param counter  : number of strips.
 *  \return int : number of vertices.
 */
int
MetricBarriolaVilenkin :: getEmbeddingVertices ( std::vector<vec3> &verts,
                                                 std::vector<int> &indices, unsigned int &numElems, unsigned int &counter )
{
  if (!verts.empty())
    verts.clear();

  if (!indices.empty())
    indices.clear();

  if (fabs(mk)>1.0)
     return 0;

  testEmbeddingParams();
  mEmb_rstep = (mEmb_rmax-mEmb_rmin)/mEmb_r_num;
  mEmb_phistep = 2.0*M_PI/mEmb_phi_num;

  numElems = int( mEmb_r_num );
  counter  = int( mEmb_phi_num ) + 1;

  int vnum;

  double x,y,z,r,phi;
  for(unsigned int k=0; k<counter; k++)
  {
    phi = k*mEmb_phistep;
    for(unsigned int j=0; j<numElems; j++)
    {
      r = mEmb_rmin + j*mEmb_rstep;
      x = r*cos(phi);
      y = r*sin(phi);

        z = sqrt(1.0-mk*mk)*r;
        verts.push_back(vec3(x,y,z));

        vnum = k*numElems+j;

        indices.push_back( vnum );
        indices.push_back( vnum+numElems );

    }
  }

  int numVerts = (int)verts.size();
  int numInds  = (int)indices.size();

  if (2*numVerts==numInds)
    return numVerts;

  return 0;
}

/*! Effective potential.
 *  \param pos : initial position.
 *  \param cdir : initial four-direction.
 *  \param type : geodesic type.
 *  \param x : abscissa value.
 *  \param val : reference to effective potential value.
 *  \return true : effective potential exists at x.
 */
bool
MetricBarriolaVilenkin :: effPotentialValue  ( const vec4 pos, const vec4 cdir ,enum_geodesic_type type, const double x, double &val )
{
  double kappa = 0.0;
  if (type==enum_geodesic_timelike)
    kappa = -mSign;

  if (x<0.0)
     return false;

  double h = mk*mk*pos[1]*pos[1]*cdir[3];
  val = 0.5*(h*h/(mk*mk*x*x)-kappa*mSpeedOfLight*mSpeedOfLight);
  return true;
}

/*! Total energy.
 *  \param pos : initial position.
 *  \param cdir : initial four-direction.
 *  \param x : abscissa value.
 *  \param val : reference to total energy value.
 *  \return true : effective potential exists at x.
 */
bool
MetricBarriolaVilenkin :: totEnergy ( const vec4 , const vec4 cdir, const double , double &val )
{
  // 1/2*k^2/c^2:
  val = 0.5*mSpeedOfLight*mSpeedOfLight*cdir[0]*cdir[0];
  return true;
}

/*!  Generate report.    
 */
bool    
MetricBarriolaVilenkin :: report ( const vec4 pos, const vec4 cdir, std::string &text )
{
  std::stringstream ss;
  ss << "Report for Barriola-Vilenkin metric\n\tcoordinate : (t,r,theta,phi)\n";
  ss << "---------------------------------------------------------------\n";
  ss << "  physical units ....................... yes\n";
  ss.precision(DEF_FIXED_REPORT_PRECISION);
  ss.setf(std::ios::fixed);

  double h2 = mk*mk*pos[1]*pos[1]*cdir[3];
  double h1 = mSpeedOfLight*mSpeedOfLight*cdir[0];
  ss << "  constant of motion ........ h1 = " << h1 << std::endl;
  ss << "  constant of motion ........ h2 = " << h2 << std::endl;

  double r = closestApproach(pos,cdir);
  ss << "  closest approach ....... r_pca = " << r << std::endl;
  text = ss.str();
  return true;
}

// ********************************* protected methods *****************************
/*!
 */
void 
MetricBarriolaVilenkin :: setStandardValues ( )
{
  mInitPos[0] = 0.0; 
  mInitPos[1] = 10.0;
  mInitPos[2] = M_PI_2;
  mInitPos[3] = 0.0;
  mInitDir[0] = 1.0;
  mInitDir[1] = 0.0;
  mInitDir[2] = 0.0;
   
  mCoordNames[0] = std::string("t");
  mCoordNames[1] = std::string("r");
  mCoordNames[2] = std::string("theta");
  mCoordNames[3] = std::string("phi"); 
}

/*! Radius of closest approach.
 * \param pos : initial position.
 * \param cdir : initial direction.
 * \return radius.
 */
double
MetricBarriolaVilenkin :: closestApproach ( const vec4 pos, const vec4 cdir )
{
   double h2 = mk*mk*pos[1]*pos[1]*cdir[3];
   double h1 = mSpeedOfLight*mSpeedOfLight*cdir[0];

   return h2*mSpeedOfLight/(mk*h1);
}

} // end namespace m4d
