// -------------------------------------------------------------------------------
/*
   m4dMetricBertottiKasner.cpp
 
  Copyright (c) 2010  Thomas Mueller
 
 
   This file is part of the m4d-library.
 
   The m4d-library is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
 
   The m4d-library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with the m4d-library.  If not, see <http://www.gnu.org/licenses/>.
 
*/
// -------------------------------------------------------------------------------

#include "m4dMetricBertottiKasner.h"

namespace m4d
{
    
#define eps 1.0e-6


/*! Standard constructor for the BertottiKasner metric.
 *
 * \param  lambda : cosmological constant.
 */
MetricBertottiKasner :: MetricBertottiKasner ( double lambda )
{
  mMetricName  = "MetricBertottiKasner";
  setCoordType(enum_coordinate_spherical);
    
  mPhysicalUnits = enum_physical_constants_geom;
  mSpeedOfLight   = 1.0;
  mGravConstant   = 1.0;
  mDielectricPerm = 1.0;
    
  addParam("lambda",lambda);
  mLambda = lambda;

  mLocTeds.push_back(enum_nat_tetrad_static);

  setStandardValues();
}

/*!
 */
MetricBertottiKasner :: ~MetricBertottiKasner()
{
}


// *********************************** public methods ******************************

/*! Calculate the contravariant metric components at position 'pos'.
 *
 *  \param pos : pointer to position.
 */
bool 
MetricBertottiKasner :: calculateMetric  ( const double* pos )
{
  double t     = pos[0];
  double theta = pos[2];

  double c = mSpeedOfLight;

  g_compts[0][0] = -c*c;
  g_compts[0][1] = 0.0;
  g_compts[0][2] = 0.0;
  g_compts[0][3] = 0.0;
  g_compts[1][0] = 0.0;
  g_compts[1][1] = exp(2.0*sqrt(mLambda)*c*t);
  g_compts[1][2] = 0.0;
  g_compts[1][3] = 0.0;
  g_compts[2][0] = 0.0;
  g_compts[2][1] = 0.0;
  g_compts[2][2] = 1.0/mLambda;
  g_compts[2][3] = 0.0;
  g_compts[3][0] = 0.0;
  g_compts[3][1] = 0.0;
  g_compts[3][2] = 0.0;
  g_compts[3][3] = sin(theta)*sin(theta)/mLambda;

  return true;
}

/*! Calculate the Christoffel symbols of the second kind at position 'pos'.
 *
 *  \param pos : pointer to position.
 */
bool 
MetricBertottiKasner :: calculateChristoffels  ( const double* pos )
{
   double t     = pos[0];
   double theta = pos[2];
  
   double c = mSpeedOfLight;
   double sl = sqrt(mLambda);
   double ct = cos(theta)/sin(theta);

   christoffel[0][0][0] = 0.0;
   christoffel[0][0][1] = 0.0;
   christoffel[0][0][2] = 0.0;
   christoffel[0][0][3] = 0.0;
   christoffel[0][1][0] = 0.0;
   christoffel[0][1][1] = c*sl;
   christoffel[0][1][2] = 0.0;
   christoffel[0][1][3] = 0.0;
   christoffel[0][2][0] = 0.0;
   christoffel[0][2][1] = 0.0;
   christoffel[0][2][2] = 0.0;
   christoffel[0][2][3] = 0.0;
   christoffel[0][3][0] = 0.0;
   christoffel[0][3][1] = 0.0;
   christoffel[0][3][2] = 0.0;
   christoffel[0][3][3] = 0.0;
   christoffel[1][0][0] = 0.0;
   christoffel[1][0][1] = c*sl;
   christoffel[1][0][2] = 0.0;
   christoffel[1][0][3] = 0.0;
   christoffel[1][1][0] = sl/c*exp(2.0*sl*c*t);
   christoffel[1][1][1] = 0.0;
   christoffel[1][1][2] = 0.0;
   christoffel[1][1][3] = 0.0;
   christoffel[1][2][0] = 0.0;
   christoffel[1][2][1] = 0.0;
   christoffel[1][2][2] = 0.0;
   christoffel[1][2][3] = 0.0;
   christoffel[1][3][0] = 0.0;
   christoffel[1][3][1] = 0.0;
   christoffel[1][3][2] = 0.0;
   christoffel[1][3][3] = 0.0;
   christoffel[2][0][0] = 0.0;
   christoffel[2][0][1] = 0.0;
   christoffel[2][0][2] = 0.0;
   christoffel[2][0][3] = 0.0;
   christoffel[2][1][0] = 0.0;
   christoffel[2][1][1] = 0.0;
   christoffel[2][1][2] = 0.0;
   christoffel[2][1][3] = 0.0;
   christoffel[2][2][0] = 0.0;
   christoffel[2][2][1] = 0.0;
   christoffel[2][2][2] = 0.0;
   christoffel[2][2][3] = 0.0;
   christoffel[2][3][0] = 0.0;
   christoffel[2][3][1] = 0.0;
   christoffel[2][3][2] = 0.0;
   christoffel[2][3][3] = ct;
   christoffel[3][0][0] = 0.0;
   christoffel[3][0][1] = 0.0;
   christoffel[3][0][2] = 0.0;
   christoffel[3][0][3] = 0.0;
   christoffel[3][1][0] = 0.0;
   christoffel[3][1][1] = 0.0;
   christoffel[3][1][2] = 0.0;
   christoffel[3][1][3] = 0.0;
   christoffel[3][2][0] = 0.0;
   christoffel[3][2][1] = 0.0;
   christoffel[3][2][2] = 0.0;
   christoffel[3][2][3] = ct;
   christoffel[3][3][0] = 0.0;
   christoffel[3][3][1] = 0.0;
   christoffel[3][3][2] = -sin(theta)*cos(theta);
   christoffel[3][3][3] = 0.0;

  return true;
}

/*! Calculate Jacobi matrix. 
 *
 *  \param pos : pointer to position.
 */
bool 
MetricBertottiKasner :: calculateChrisD ( const double* pos )
{
   double t     = pos[0];
   double theta = pos[2];

   double c = mSpeedOfLight;
   double sl = sqrt(mLambda);

   chrisD[0][0][0][0] = 0.0;
   chrisD[0][0][0][1] = 0.0;
   chrisD[0][0][0][2] = 0.0;
   chrisD[0][0][0][3] = 0.0;
   chrisD[0][0][1][0] = 0.0;
   chrisD[0][0][1][1] = 0.0;
   chrisD[0][0][1][2] = 0.0;
   chrisD[0][0][1][3] = 0.0;
   chrisD[0][0][2][0] = 0.0;
   chrisD[0][0][2][1] = 0.0;
   chrisD[0][0][2][2] = 0.0;
   chrisD[0][0][2][3] = 0.0;
   chrisD[0][0][3][0] = 0.0;
   chrisD[0][0][3][1] = 0.0;
   chrisD[0][0][3][2] = 0.0;
   chrisD[0][0][3][3] = 0.0;
   chrisD[0][1][0][0] = 0.0;
   chrisD[0][1][0][1] = 0.0;
   chrisD[0][1][0][2] = 0.0;
   chrisD[0][1][0][3] = 0.0;
   chrisD[0][1][1][0] = 0.0;
   chrisD[0][1][1][1] = 0.0;
   chrisD[0][1][1][2] = 0.0;
   chrisD[0][1][1][3] = 0.0;
   chrisD[0][1][2][0] = 0.0;
   chrisD[0][1][2][1] = 0.0;
   chrisD[0][1][2][2] = 0.0;
   chrisD[0][1][2][3] = 0.0;
   chrisD[0][1][3][0] = 0.0;
   chrisD[0][1][3][1] = 0.0;
   chrisD[0][1][3][2] = 0.0;
   chrisD[0][1][3][3] = 0.0;
   chrisD[0][2][0][0] = 0.0;
   chrisD[0][2][0][1] = 0.0;
   chrisD[0][2][0][2] = 0.0;
   chrisD[0][2][0][3] = 0.0;
   chrisD[0][2][1][0] = 0.0;
   chrisD[0][2][1][1] = 0.0;
   chrisD[0][2][1][2] = 0.0;
   chrisD[0][2][1][3] = 0.0;
   chrisD[0][2][2][0] = 0.0;
   chrisD[0][2][2][1] = 0.0;
   chrisD[0][2][2][2] = 0.0;
   chrisD[0][2][2][3] = 0.0;
   chrisD[0][2][3][0] = 0.0;
   chrisD[0][2][3][1] = 0.0;
   chrisD[0][2][3][2] = 0.0;
   chrisD[0][2][3][3] = 0.0;
   chrisD[0][3][0][0] = 0.0;
   chrisD[0][3][0][1] = 0.0;
   chrisD[0][3][0][2] = 0.0;
   chrisD[0][3][0][3] = 0.0;
   chrisD[0][3][1][0] = 0.0;
   chrisD[0][3][1][1] = 0.0;
   chrisD[0][3][1][2] = 0.0;
   chrisD[0][3][1][3] = 0.0;
   chrisD[0][3][2][0] = 0.0;
   chrisD[0][3][2][1] = 0.0;
   chrisD[0][3][2][2] = 0.0;
   chrisD[0][3][2][3] = 0.0;
   chrisD[0][3][3][0] = 0.0;
   chrisD[0][3][3][1] = 0.0;
   chrisD[0][3][3][2] = 0.0;
   chrisD[0][3][3][3] = 0.0;
   chrisD[1][0][0][0] = 0.0;
   chrisD[1][0][0][1] = 0.0;
   chrisD[1][0][0][2] = 0.0;
   chrisD[1][0][0][3] = 0.0;
   chrisD[1][0][1][0] = 0.0;
   chrisD[1][0][1][1] = 0.0;
   chrisD[1][0][1][2] = 0.0;
   chrisD[1][0][1][3] = 0.0;
   chrisD[1][0][2][0] = 0.0;
   chrisD[1][0][2][1] = 0.0;
   chrisD[1][0][2][2] = 0.0;
   chrisD[1][0][2][3] = 0.0;
   chrisD[1][0][3][0] = 0.0;
   chrisD[1][0][3][1] = 0.0;
   chrisD[1][0][3][2] = 0.0;
   chrisD[1][0][3][3] = 0.0;
   chrisD[1][1][0][0] = 2.0*mLambda*exp(2.0*sl*c*t);
   chrisD[1][1][0][1] = 0.0;
   chrisD[1][1][0][2] = 0.0;
   chrisD[1][1][0][3] = 0.0;
   chrisD[1][1][1][0] = 0.0;
   chrisD[1][1][1][1] = 0.0;
   chrisD[1][1][1][2] = 0.0;
   chrisD[1][1][1][3] = 0.0;
   chrisD[1][1][2][0] = 0.0;
   chrisD[1][1][2][1] = 0.0;
   chrisD[1][1][2][2] = 0.0;
   chrisD[1][1][2][3] = 0.0;
   chrisD[1][1][3][0] = 0.0;
   chrisD[1][1][3][1] = 0.0;
   chrisD[1][1][3][2] = 0.0;
   chrisD[1][1][3][3] = 0.0;
   chrisD[1][2][0][0] = 0.0;
   chrisD[1][2][0][1] = 0.0;
   chrisD[1][2][0][2] = 0.0;
   chrisD[1][2][0][3] = 0.0;
   chrisD[1][2][1][0] = 0.0;
   chrisD[1][2][1][1] = 0.0;
   chrisD[1][2][1][2] = 0.0;
   chrisD[1][2][1][3] = 0.0;
   chrisD[1][2][2][0] = 0.0;
   chrisD[1][2][2][1] = 0.0;
   chrisD[1][2][2][2] = 0.0;
   chrisD[1][2][2][3] = 0.0;
   chrisD[1][2][3][0] = 0.0;
   chrisD[1][2][3][1] = 0.0;
   chrisD[1][2][3][2] = 0.0;
   chrisD[1][2][3][3] = 0.0;
   chrisD[1][3][0][0] = 0.0;
   chrisD[1][3][0][1] = 0.0;
   chrisD[1][3][0][2] = 0.0;
   chrisD[1][3][0][3] = 0.0;
   chrisD[1][3][1][0] = 0.0;
   chrisD[1][3][1][1] = 0.0;
   chrisD[1][3][1][2] = 0.0;
   chrisD[1][3][1][3] = 0.0;
   chrisD[1][3][2][0] = 0.0;
   chrisD[1][3][2][1] = 0.0;
   chrisD[1][3][2][2] = 0.0;
   chrisD[1][3][2][3] = 0.0;
   chrisD[1][3][3][0] = 0.0;
   chrisD[1][3][3][1] = 0.0;
   chrisD[1][3][3][2] = 0.0;
   chrisD[1][3][3][3] = 0.0;
   chrisD[2][0][0][0] = 0.0;
   chrisD[2][0][0][1] = 0.0;
   chrisD[2][0][0][2] = 0.0;
   chrisD[2][0][0][3] = 0.0;
   chrisD[2][0][1][0] = 0.0;
   chrisD[2][0][1][1] = 0.0;
   chrisD[2][0][1][2] = 0.0;
   chrisD[2][0][1][3] = 0.0;
   chrisD[2][0][2][0] = 0.0;
   chrisD[2][0][2][1] = 0.0;
   chrisD[2][0][2][2] = 0.0;
   chrisD[2][0][2][3] = 0.0;
   chrisD[2][0][3][0] = 0.0;
   chrisD[2][0][3][1] = 0.0;
   chrisD[2][0][3][2] = 0.0;
   chrisD[2][0][3][3] = 0.0;
   chrisD[2][1][0][0] = 0.0;
   chrisD[2][1][0][1] = 0.0;
   chrisD[2][1][0][2] = 0.0;
   chrisD[2][1][0][3] = 0.0;
   chrisD[2][1][1][0] = 0.0;
   chrisD[2][1][1][1] = 0.0;
   chrisD[2][1][1][2] = 0.0;
   chrisD[2][1][1][3] = 0.0;
   chrisD[2][1][2][0] = 0.0;
   chrisD[2][1][2][1] = 0.0;
   chrisD[2][1][2][2] = 0.0;
   chrisD[2][1][2][3] = 0.0;
   chrisD[2][1][3][0] = 0.0;
   chrisD[2][1][3][1] = 0.0;
   chrisD[2][1][3][2] = 0.0;
   chrisD[2][1][3][3] = 0.0;
   chrisD[2][2][0][0] = 0.0;
   chrisD[2][2][0][1] = 0.0;
   chrisD[2][2][0][2] = 0.0;
   chrisD[2][2][0][3] = 0.0;
   chrisD[2][2][1][0] = 0.0;
   chrisD[2][2][1][1] = 0.0;
   chrisD[2][2][1][2] = 0.0;
   chrisD[2][2][1][3] = 0.0;
   chrisD[2][2][2][0] = 0.0;
   chrisD[2][2][2][1] = 0.0;
   chrisD[2][2][2][2] = 0.0;
   chrisD[2][2][2][3] = 0.0;
   chrisD[2][2][3][0] = 0.0;
   chrisD[2][2][3][1] = 0.0;
   chrisD[2][2][3][2] = 0.0;
   chrisD[2][2][3][3] = 0.0;
   chrisD[2][3][0][0] = 0.0;
   chrisD[2][3][0][1] = 0.0;
   chrisD[2][3][0][2] = 0.0;
   chrisD[2][3][0][3] = 0.0;
   chrisD[2][3][1][0] = 0.0;
   chrisD[2][3][1][1] = 0.0;
   chrisD[2][3][1][2] = 0.0;
   chrisD[2][3][1][3] = 0.0;
   chrisD[2][3][2][0] = 0.0;
   chrisD[2][3][2][1] = 0.0;
   chrisD[2][3][2][2] = 0.0;
   chrisD[2][3][2][3] = 0.0;
   chrisD[2][3][3][0] = 0.0;
   chrisD[2][3][3][1] = 0.0;
   chrisD[2][3][3][2] = -pow(sin(theta),-2.0);
   chrisD[2][3][3][3] = 0.0;
   chrisD[3][0][0][0] = 0.0;
   chrisD[3][0][0][1] = 0.0;
   chrisD[3][0][0][2] = 0.0;
   chrisD[3][0][0][3] = 0.0;
   chrisD[3][0][1][0] = 0.0;
   chrisD[3][0][1][1] = 0.0;
   chrisD[3][0][1][2] = 0.0;
   chrisD[3][0][1][3] = 0.0;
   chrisD[3][0][2][0] = 0.0;
   chrisD[3][0][2][1] = 0.0;
   chrisD[3][0][2][2] = 0.0;
   chrisD[3][0][2][3] = 0.0;
   chrisD[3][0][3][0] = 0.0;
   chrisD[3][0][3][1] = 0.0;
   chrisD[3][0][3][2] = 0.0;
   chrisD[3][0][3][3] = 0.0;
   chrisD[3][1][0][0] = 0.0;
   chrisD[3][1][0][1] = 0.0;
   chrisD[3][1][0][2] = 0.0;
   chrisD[3][1][0][3] = 0.0;
   chrisD[3][1][1][0] = 0.0;
   chrisD[3][1][1][1] = 0.0;
   chrisD[3][1][1][2] = 0.0;
   chrisD[3][1][1][3] = 0.0;
   chrisD[3][1][2][0] = 0.0;
   chrisD[3][1][2][1] = 0.0;
   chrisD[3][1][2][2] = 0.0;
   chrisD[3][1][2][3] = 0.0;
   chrisD[3][1][3][0] = 0.0;
   chrisD[3][1][3][1] = 0.0;
   chrisD[3][1][3][2] = 0.0;
   chrisD[3][1][3][3] = 0.0;
   chrisD[3][2][0][0] = 0.0;
   chrisD[3][2][0][1] = 0.0;
   chrisD[3][2][0][2] = 0.0;
   chrisD[3][2][0][3] = 0.0;
   chrisD[3][2][1][0] = 0.0;
   chrisD[3][2][1][1] = 0.0;
   chrisD[3][2][1][2] = 0.0;
   chrisD[3][2][1][3] = 0.0;
   chrisD[3][2][2][0] = 0.0;
   chrisD[3][2][2][1] = 0.0;
   chrisD[3][2][2][2] = 0.0;
   chrisD[3][2][2][3] = 0.0;
   chrisD[3][2][3][0] = 0.0;
   chrisD[3][2][3][1] = 0.0;
   chrisD[3][2][3][2] = -pow(sin(theta),-2.0);
   chrisD[3][2][3][3] = 0.0;
   chrisD[3][3][0][0] = 0.0;
   chrisD[3][3][0][1] = 0.0;
   chrisD[3][3][0][2] = 0.0;
   chrisD[3][3][0][3] = 0.0;
   chrisD[3][3][1][0] = 0.0;
   chrisD[3][3][1][1] = 0.0;
   chrisD[3][3][1][2] = 0.0;
   chrisD[3][3][1][3] = 0.0;
   chrisD[3][3][2][0] = 0.0;
   chrisD[3][3][2][1] = 0.0;
   chrisD[3][3][2][2] = -cos(2.0*theta);
   chrisD[3][3][2][3] = 0.0;
   chrisD[3][3][3][0] = 0.0;
   chrisD[3][3][3][1] = 0.0;
   chrisD[3][3][3][2] = 0.0;
   chrisD[3][3][3][3] = 0.0;

  return true;
}


/*! Transform local 4-direction to coordinate 4-direction.
 *
 *  \param  pos  :  pointer to position array.
 *  \param  ldir :  pointer to local direction array.
 *  \param  dir  :  pointer to calculated coordinate direction array.
 *  \param  type :  type of tetrad.
 */
void
MetricBertottiKasner :: localToCoord   ( const double* pos, const double* ldir, double* dir,
                                         enum_nat_tetrad_type   )
{ 
  double t     = pos[0];
  double theta = pos[2];
  
  double c = mSpeedOfLight;
  double sl = sqrt(mLambda);

  dir[0] = ldir[0]/c;
  dir[1] = ldir[1]*exp(-sl*c*t);
  dir[2] = ldir[2]*sl;
  dir[3] = ldir[3]*sl/sin(theta);
}

/*! Transform coordinate 4-direction to local 4-direction.
 *
 *  \param  pos  :  pointer to position array.
 *  \param  cdir :  pointer to coordinate direction.
 *  \param  ldir :  pointer to calculated local direction array.
 *  \param  type :  type of tetrad.
 */
void 
MetricBertottiKasner :: coordToLocal   ( const double* pos, const double* cdir, double* ldir,
                                         enum_nat_tetrad_type   )
{
  double t     = pos[0];
  double theta = pos[2];
 
  double sl = sqrt(mLambda);
     
  ldir[0] = cdir[0];
  ldir[1] = cdir[1]*exp(sl*t);
  ldir[2] = cdir[2]/sl;
  ldir[3] = cdir[3]/sl*sin(theta);
}


/*! Test break condition.
 *
 *  \param pos    : pointer to position array.
 *  \return true  : radial position r < 0.0 or  r^2<=(1.0+eps)*rs^2.
 *  \return false : position is valid.
 */
bool 
MetricBertottiKasner :: breakCondition ( const double* )
{
  return false;
}


/*! Tests whether the constraint equation is fulfilled.
 *
 *  The constraint equation for lightlike and timelike geodesics reads:  
 \verbatim
     sum = g_{\mu\nu} dot(x)^{\mu} dot(x)^{\nu} - kappa c^2 = 0.
 \endverbatim
 *  \param  y[]   : pointer to position and direction coordinates.
 *  \param  kappa : timelike (-1.0), lightlike (0.0).
 *  \return double : sum.
 */
double 
MetricBertottiKasner :: testConstraint ( const double y[], const double kappa )
{
  double t     = y[0];
  double theta = y[2];

  double dt     = y[4];
  double dr     = y[5];
  double dtheta = y[6];
  double dphi   = y[7];

  double sum = -kappa*mSign;
  sum += -dt*dt + exp(2.0*sqrt(mLambda)*t)*dr*dr + (dtheta*dtheta + sin(theta)*sin(theta)*dphi*dphi)/mLambda;
  return sum;
}

/*! Set parameter 'pName' to 'val'.
 *
 *  Set 'mass' parameter and adjust Schwarzschild radius  rs=2GM/c^2.
 *  'charge' represents the charge of the black hole.
 */
bool 
MetricBertottiKasner :: setParam ( std::string pName, double val )
{
  Metric::setParam(pName,val);
  
  if (pName=="lambda")
  {
    mLambda = val;
  }
  return true;
}


/*! Generate report.
 * \param pos : initial position.
 * \param cdir : initial coordinate direction.
 * \param text : reference to report text.
 */
bool    
MetricBertottiKasner :: report ( const vec4 pos, const vec4 cdir, std::string &text )
{
  std::stringstream ss;
  ss << "Report for the  BertottiKasner metric\n\tcoordinates : (t,r,theta,phi)\n";
  ss << "---------------------------------------------------------------\n";
  ss.precision(DEF_FIXED_REPORT_PRECISION);
  ss.setf(std::ios::fixed);
  ss << "  Lambda ........................... " << mLambda << std::endl;

  double h1 = exp(2.0*sqrt(mLambda)*pos[0])*cdir[1];
  double h2 = cdir[3]/mLambda;
  ss << " constant of motion             h1 = " << h1 << std::endl;
  ss << " constant of motion             h2 = " << h2 << std::endl;
  text = ss.str();
  return true;
}

// ********************************* protected methods *****************************
/*!
 */
void 
MetricBertottiKasner :: setStandardValues( )
{
  mInitPos[0] = -1.0;
  mInitPos[1] = 1.0;
  mInitPos[2] = M_PI_2;
  mInitPos[3] = 0.0;
  mInitDir[0] = 1.0;
  mInitDir[1] = 0.0;
  mInitDir[2] = 0.0;
   
  mCoordNames[0] = std::string("t");
  mCoordNames[1] = std::string("r");
  mCoordNames[2] = std::string("theta");
  mCoordNames[3] = std::string("phi"); 
}

} // end namespace m4d
