// -------------------------------------------------------------------------------
/*
   m4dMetricExtremeReissnerNordstromDihole.cpp
 
  Copyright (c) 2009  Thomas Mueller, Frank Grave
 
 
   This file is part of the m4d-library.
 
   The m4d-library is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
 
   The m4d-library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with the m4d-library.  If not, see <http://www.gnu.org/licenses/>.
 
*/
// -------------------------------------------------------------------------------

#include "m4dMetricExtremeReissnerNordstromDihole.h"

namespace m4d
{

#define eps 1.0e-6


/*! Standard constructor for the Kottler metric.
 *
 * \param  mass1 : mass of the first black hole.
 * \param  mass2 : mass of the second black hole.
 */
MetricExtremeReissnerNordstromDihole :: MetricExtremeReissnerNordstromDihole ( double mass1, double mass2 )
{
  mMetricName  = "ExtremeReissnerNordstromDihole";
  setCoordType(enum_coordinate_cartesian);
    
  mPhysicalUnits = enum_physical_constants_geom;
  mSpeedOfLight = 1.0;
  mGravConstant = 1.0;
  
  addParam("m1",mass1);
  addParam("m2",mass2);
  mM1 = mass1;
  mM2 = mass2;
  
  mDrawTypes.push_back(enum_draw_twoplusone);

  mLocTeds.push_back(enum_nat_tetrad_static);
  setStandardValues();
}

MetricExtremeReissnerNordstromDihole :: ~MetricExtremeReissnerNordstromDihole()
{
}


// *********************************** public methods ******************************
/*! Calculate the contravariant metric components at position 'pos'.
 *
 *  \param pos : pointer to position.
 */
bool 
MetricExtremeReissnerNordstromDihole :: calculateMetric  ( const double* pos )
{
  double t1 = calc_U(pos);  // U(x,y,z);
  double t2 = t1*t1;
	  
      g_compts[0][0] = -1/t2;
      g_compts[0][1] = 0.0;
      g_compts[0][2] = 0.0;
      g_compts[0][3] = 0.0;
      g_compts[1][0] = 0.0;
      g_compts[1][1] = t2;
      g_compts[1][2] = 0.0;
      g_compts[1][3] = 0.0;
      g_compts[2][0] = 0.0;
      g_compts[2][1] = 0.0;
      g_compts[2][2] = t2;
      g_compts[2][3] = 0.0;
      g_compts[3][0] = 0.0;
      g_compts[3][1] = 0.0;
      g_compts[3][2] = 0.0;
      g_compts[3][3] = t2;
	  
  return true;
}

/*! Calculate the Christoffel symbols of the second kind at position 'pos'.
 *
 *  \param pos : pointer to position.
 */
bool 
MetricExtremeReissnerNordstromDihole :: calculateChristoffels  ( const double* pos )
{
  double t1 = calc_U(pos);  // U(x,y,z);
  double Ux,Uy,Uz;
  calc_dU(pos,Ux,Uy,Uz);
  
  double t2 = t1*t1;
  double t3 = t2*t2;
  double t5 = 1/t3/t1;
  double t6 = Ux;   // diff(U(x,y,z),x);
  double t8 = Uy;   // diff(U(x,y,z),y);
  double t10 = Uz;  // diff(U(x,y,z),z);
  double t12 = 1/t1;
  double t13 = t12*t6;
  double t14 = t12*t8;
  double t15 = t12*t10;
	  
      christoffel[0][0][0] = 0.0;
      christoffel[0][0][1] = -t5*t6;
      christoffel[0][0][2] = -t5*t8;
      christoffel[0][0][3] = -t5*t10;
      christoffel[0][1][0] = -t13;
      christoffel[0][1][1] = 0.0;
      christoffel[0][1][2] = 0.0;
      christoffel[0][1][3] = 0.0;
      christoffel[0][2][0] = -t14;
      christoffel[0][2][1] = 0.0;
      christoffel[0][2][2] = 0.0;
      christoffel[0][2][3] = 0.0;
      christoffel[0][3][0] = -t15;
      christoffel[0][3][1] = 0.0;
      christoffel[0][3][2] = 0.0;
      christoffel[0][3][3] = 0.0;
      christoffel[1][0][0] = -t13;
      christoffel[1][0][1] = 0.0;
      christoffel[1][0][2] = 0.0;
      christoffel[1][0][3] = 0.0;
      christoffel[1][1][0] = 0.0;
      christoffel[1][1][1] = t13;
      christoffel[1][1][2] = -t14;
      christoffel[1][1][3] = -t15;
      christoffel[1][2][0] = 0.0;
      christoffel[1][2][1] = t14;
      christoffel[1][2][2] = t13;
      christoffel[1][2][3] = 0.0;
      christoffel[1][3][0] = 0.0;
      christoffel[1][3][1] = t15;
      christoffel[1][3][2] = 0.0;
      christoffel[1][3][3] = t13;
      christoffel[2][0][0] = -t14;
      christoffel[2][0][1] = 0.0;
      christoffel[2][0][2] = 0.0;
      christoffel[2][0][3] = 0.0;
      christoffel[2][1][0] = 0.0;
      christoffel[2][1][1] = t14;
      christoffel[2][1][2] = t13;
      christoffel[2][1][3] = 0.0;
      christoffel[2][2][0] = 0.0;
      christoffel[2][2][1] = -t13;
      christoffel[2][2][2] = t14;
      christoffel[2][2][3] = -t15;
      christoffel[2][3][0] = 0.0;
      christoffel[2][3][1] = 0.0;
      christoffel[2][3][2] = t15;
      christoffel[2][3][3] = t14;
      christoffel[3][0][0] = -t15;
      christoffel[3][0][1] = 0.0;
      christoffel[3][0][2] = 0.0;
      christoffel[3][0][3] = 0.0;
      christoffel[3][1][0] = 0.0;
      christoffel[3][1][1] = t15;
      christoffel[3][1][2] = 0.0;
      christoffel[3][1][3] = t13;
      christoffel[3][2][0] = 0.0;
      christoffel[3][2][1] = 0.0;
      christoffel[3][2][2] = t15;
      christoffel[3][2][3] = t14;
      christoffel[3][3][0] = 0.0;
      christoffel[3][3][1] = -t13;
      christoffel[3][3][2] = -t14;
      christoffel[3][3][3] = t15;
	  
  return true;
}

/*! Calculate Jacobi matrix. 
 *
 *  \param pos : pointer to position.
 */
bool 
MetricExtremeReissnerNordstromDihole :: calculateChrisD ( const double* pos )
{
  double U = calc_U(pos);
  double Ux,Uy,Uz,Uxx,Uyy,Uzz,Uxy,Uxz,Uyz;
  calc_dU(pos,Ux,Uy,Uz);
  calc_ddU(pos,Uxx,Uyy,Uzz,Uxy,Uxz,Uyz);
  
  double t1 = Ux;      //diff(U(x,y,z),x);
  double t2 = t1*t1;
  double t4 = Uxx;     // diff(diff(U(x,y,z),x),x);
  double t5 = U;       // U(x,y,z);
  double t6 = t4*t5;
  double t8 = t5*t5;
  double t9 = t8*t8;
  double t11 = 1/t9/t8;
  double t13 = Uy;     // diff(U(x,y,z),y);
  double t14 = t1*t13;
  double t16 = Uxy;    // diff(diff(U(x,y,z),x),y);
  double t17 = t16*t5;
  double t19 = (-5.0*t14+t17)*t11;
  double t20 = Uz;     // diff(U(x,y,z),z);
  double t21 = t1*t20;
  double t23 = Uxz;    // diff(diff(U(x,y,z),x),z);
  double t24 = t23*t5;
  double t26 = (-5.0*t21+t24)*t11;
  double t27 = t13*t13;
  double t29 = Uyy;    // diff(diff(U(x,y,z),y),y);
  double t30 = t29*t5;
  double t33 = t13*t20;
  double t35 = Uyz;    // diff(diff(U(x,y,z),y),z);
  double t36 = t35*t5;
  double t38 = (-5.0*t33+t36)*t11;
  double t39 = t20*t20;
  double t41 = Uzz;    // diff(diff(U(x,y,z),z),z);
  double t42 = t41*t5;
  double t46 = 1/t8;
  double t47 = (-t2+t6)*t46;
  double t49 = (-t14+t17)*t46;
  double t51 = (-t21+t24)*t46;
  double t53 = (-t27+t30)*t46;
  double t55 = (-t33+t36)*t46;
  double t57 = (-t39+t42)*t46;
  
      chrisD[0][0][0][0] = 0.0;
      chrisD[0][0][0][1] = 0.0;
      chrisD[0][0][0][2] = 0.0;
      chrisD[0][0][0][3] = 0.0;
      chrisD[0][0][1][0] = 0.0;
      chrisD[0][0][1][1] = -(-5.0*t2+t6)*t11;
      chrisD[0][0][1][2] = -t19;
      chrisD[0][0][1][3] = -t26;
      chrisD[0][0][2][0] = 0.0;
      chrisD[0][0][2][1] = -t19;
      chrisD[0][0][2][2] = -(-5.0*t27+t30)*t11;
      chrisD[0][0][2][3] = -t38;
      chrisD[0][0][3][0] = 0.0;
      chrisD[0][0][3][1] = -t26;
      chrisD[0][0][3][2] = -t38;
      chrisD[0][0][3][3] = -(-5.0*t39+t42)*t11;
      chrisD[0][1][0][0] = 0.0;
      chrisD[0][1][0][1] = -t47;
      chrisD[0][1][0][2] = -t49;
      chrisD[0][1][0][3] = -t51;
      chrisD[0][1][1][0] = 0.0;
      chrisD[0][1][1][1] = 0.0;
      chrisD[0][1][1][2] = 0.0;
      chrisD[0][1][1][3] = 0.0;
      chrisD[0][1][2][0] = 0.0;
      chrisD[0][1][2][1] = 0.0;
      chrisD[0][1][2][2] = 0.0;
      chrisD[0][1][2][3] = 0.0;
      chrisD[0][1][3][0] = 0.0;
      chrisD[0][1][3][1] = 0.0;
      chrisD[0][1][3][2] = 0.0;
      chrisD[0][1][3][3] = 0.0;
      chrisD[0][2][0][0] = 0.0;
      chrisD[0][2][0][1] = -t49;
      chrisD[0][2][0][2] = -t53;
      chrisD[0][2][0][3] = -t55;
      chrisD[0][2][1][0] = 0.0;
      chrisD[0][2][1][1] = 0.0;
      chrisD[0][2][1][2] = 0.0;
      chrisD[0][2][1][3] = 0.0;
      chrisD[0][2][2][0] = 0.0;
      chrisD[0][2][2][1] = 0.0;
      chrisD[0][2][2][2] = 0.0;
      chrisD[0][2][2][3] = 0.0;
      chrisD[0][2][3][0] = 0.0;
      chrisD[0][2][3][1] = 0.0;
      chrisD[0][2][3][2] = 0.0;
      chrisD[0][2][3][3] = 0.0;
      chrisD[0][3][0][0] = 0.0;
      chrisD[0][3][0][1] = -t51;
      chrisD[0][3][0][2] = -t55;
      chrisD[0][3][0][3] = -t57;
      chrisD[0][3][1][0] = 0.0;
      chrisD[0][3][1][1] = 0.0;
      chrisD[0][3][1][2] = 0.0;
      chrisD[0][3][1][3] = 0.0;
      chrisD[0][3][2][0] = 0.0;
      chrisD[0][3][2][1] = 0.0;
      chrisD[0][3][2][2] = 0.0;
      chrisD[0][3][2][3] = 0.0;
      chrisD[0][3][3][0] = 0.0;
      chrisD[0][3][3][1] = 0.0;
      chrisD[0][3][3][2] = 0.0;
      chrisD[0][3][3][3] = 0.0;
      chrisD[1][0][0][0] = 0.0;
      chrisD[1][0][0][1] = -t47;
      chrisD[1][0][0][2] = -t49;
      chrisD[1][0][0][3] = -t51;
      chrisD[1][0][1][0] = 0.0;
      chrisD[1][0][1][1] = 0.0;
      chrisD[1][0][1][2] = 0.0;
      chrisD[1][0][1][3] = 0.0;
      chrisD[1][0][2][0] = 0.0;
      chrisD[1][0][2][1] = 0.0;
      chrisD[1][0][2][2] = 0.0;
      chrisD[1][0][2][3] = 0.0;
      chrisD[1][0][3][0] = 0.0;
      chrisD[1][0][3][1] = 0.0;
      chrisD[1][0][3][2] = 0.0;
      chrisD[1][0][3][3] = 0.0;
      chrisD[1][1][0][0] = 0.0;
      chrisD[1][1][0][1] = 0.0;
      chrisD[1][1][0][2] = 0.0;
      chrisD[1][1][0][3] = 0.0;
      chrisD[1][1][1][0] = 0.0;
      chrisD[1][1][1][1] = t47;
      chrisD[1][1][1][2] = t49;
      chrisD[1][1][1][3] = t51;
      chrisD[1][1][2][0] = 0.0;
      chrisD[1][1][2][1] = -t49;
      chrisD[1][1][2][2] = -t53;
      chrisD[1][1][2][3] = -t55;
      chrisD[1][1][3][0] = 0.0;
      chrisD[1][1][3][1] = -t51;
      chrisD[1][1][3][2] = -t55;
      chrisD[1][1][3][3] = -t57;
      chrisD[1][2][0][0] = 0.0;
      chrisD[1][2][0][1] = 0.0;
      chrisD[1][2][0][2] = 0.0;
      chrisD[1][2][0][3] = 0.0;
      chrisD[1][2][1][0] = 0.0;
      chrisD[1][2][1][1] = t49;
      chrisD[1][2][1][2] = t53;
      chrisD[1][2][1][3] = t55;
      chrisD[1][2][2][0] = 0.0;
      chrisD[1][2][2][1] = t47;
      chrisD[1][2][2][2] = t49;
      chrisD[1][2][2][3] = t51;
      chrisD[1][2][3][0] = 0.0;
      chrisD[1][2][3][1] = 0.0;
      chrisD[1][2][3][2] = 0.0;
      chrisD[1][2][3][3] = 0.0;
      chrisD[1][3][0][0] = 0.0;
      chrisD[1][3][0][1] = 0.0;
      chrisD[1][3][0][2] = 0.0;
      chrisD[1][3][0][3] = 0.0;
      chrisD[1][3][1][0] = 0.0;
      chrisD[1][3][1][1] = t51;
      chrisD[1][3][1][2] = t55;
      chrisD[1][3][1][3] = t57;
      chrisD[1][3][2][0] = 0.0;
      chrisD[1][3][2][1] = 0.0;
      chrisD[1][3][2][2] = 0.0;
      chrisD[1][3][2][3] = 0.0;
      chrisD[1][3][3][0] = 0.0;
      chrisD[1][3][3][1] = t47;
      chrisD[1][3][3][2] = t49;
      chrisD[1][3][3][3] = t51;
      chrisD[2][0][0][0] = 0.0;
      chrisD[2][0][0][1] = -t49;
      chrisD[2][0][0][2] = -t53;
      chrisD[2][0][0][3] = -t55;
      chrisD[2][0][1][0] = 0.0;
      chrisD[2][0][1][1] = 0.0;
      chrisD[2][0][1][2] = 0.0;
      chrisD[2][0][1][3] = 0.0;
      chrisD[2][0][2][0] = 0.0;
      chrisD[2][0][2][1] = 0.0;
      chrisD[2][0][2][2] = 0.0;
      chrisD[2][0][2][3] = 0.0;
      chrisD[2][0][3][0] = 0.0;
      chrisD[2][0][3][1] = 0.0;
      chrisD[2][0][3][2] = 0.0;
      chrisD[2][0][3][3] = 0.0;
      chrisD[2][1][0][0] = 0.0;
      chrisD[2][1][0][1] = 0.0;
      chrisD[2][1][0][2] = 0.0;
      chrisD[2][1][0][3] = 0.0;
      chrisD[2][1][1][0] = 0.0;
      chrisD[2][1][1][1] = t49;
      chrisD[2][1][1][2] = t53;
      chrisD[2][1][1][3] = t55;
      chrisD[2][1][2][0] = 0.0;
      chrisD[2][1][2][1] = t47;
      chrisD[2][1][2][2] = t49;
      chrisD[2][1][2][3] = t51;
      chrisD[2][1][3][0] = 0.0;
      chrisD[2][1][3][1] = 0.0;
      chrisD[2][1][3][2] = 0.0;
      chrisD[2][1][3][3] = 0.0;
      chrisD[2][2][0][0] = 0.0;
      chrisD[2][2][0][1] = 0.0;
      chrisD[2][2][0][2] = 0.0;
      chrisD[2][2][0][3] = 0.0;
      chrisD[2][2][1][0] = 0.0;
      chrisD[2][2][1][1] = -t47;
      chrisD[2][2][1][2] = -t49;
      chrisD[2][2][1][3] = -t51;
      chrisD[2][2][2][0] = 0.0;
      chrisD[2][2][2][1] = t49;
      chrisD[2][2][2][2] = t53;
      chrisD[2][2][2][3] = t55;
      chrisD[2][2][3][0] = 0.0;
      chrisD[2][2][3][1] = -t51;
      chrisD[2][2][3][2] = -t55;
      chrisD[2][2][3][3] = -t57;
      chrisD[2][3][0][0] = 0.0;
      chrisD[2][3][0][1] = 0.0;
      chrisD[2][3][0][2] = 0.0;
      chrisD[2][3][0][3] = 0.0;
      chrisD[2][3][1][0] = 0.0;
      chrisD[2][3][1][1] = 0.0;
      chrisD[2][3][1][2] = 0.0;
      chrisD[2][3][1][3] = 0.0;
      chrisD[2][3][2][0] = 0.0;
      chrisD[2][3][2][1] = t51;
      chrisD[2][3][2][2] = t55;
      chrisD[2][3][2][3] = t57;
      chrisD[2][3][3][0] = 0.0;
      chrisD[2][3][3][1] = t49;
      chrisD[2][3][3][2] = t53;
      chrisD[2][3][3][3] = t55;
      chrisD[3][0][0][0] = 0.0;
      chrisD[3][0][0][1] = -t51;
      chrisD[3][0][0][2] = -t55;
      chrisD[3][0][0][3] = -t57;
      chrisD[3][0][1][0] = 0.0;
      chrisD[3][0][1][1] = 0.0;
      chrisD[3][0][1][2] = 0.0;
      chrisD[3][0][1][3] = 0.0;
      chrisD[3][0][2][0] = 0.0;
      chrisD[3][0][2][1] = 0.0;
      chrisD[3][0][2][2] = 0.0;
      chrisD[3][0][2][3] = 0.0;
      chrisD[3][0][3][0] = 0.0;
      chrisD[3][0][3][1] = 0.0;
      chrisD[3][0][3][2] = 0.0;
      chrisD[3][0][3][3] = 0.0;
      chrisD[3][1][0][0] = 0.0;
      chrisD[3][1][0][1] = 0.0;
      chrisD[3][1][0][2] = 0.0;
      chrisD[3][1][0][3] = 0.0;
      chrisD[3][1][1][0] = 0.0;
      chrisD[3][1][1][1] = t51;
      chrisD[3][1][1][2] = t55;
      chrisD[3][1][1][3] = t57;
      chrisD[3][1][2][0] = 0.0;
      chrisD[3][1][2][1] = 0.0;
      chrisD[3][1][2][2] = 0.0;
      chrisD[3][1][2][3] = 0.0;
      chrisD[3][1][3][0] = 0.0;
      chrisD[3][1][3][1] = t47;
      chrisD[3][1][3][2] = t49;
      chrisD[3][1][3][3] = t51;
      chrisD[3][2][0][0] = 0.0;
      chrisD[3][2][0][1] = 0.0;
      chrisD[3][2][0][2] = 0.0;
      chrisD[3][2][0][3] = 0.0;
      chrisD[3][2][1][0] = 0.0;
      chrisD[3][2][1][1] = 0.0;
      chrisD[3][2][1][2] = 0.0;
      chrisD[3][2][1][3] = 0.0;
      chrisD[3][2][2][0] = 0.0;
      chrisD[3][2][2][1] = t51;
      chrisD[3][2][2][2] = t55;
      chrisD[3][2][2][3] = t57;
      chrisD[3][2][3][0] = 0.0;
      chrisD[3][2][3][1] = t49;
      chrisD[3][2][3][2] = t53;
      chrisD[3][2][3][3] = t55;
      chrisD[3][3][0][0] = 0.0;
      chrisD[3][3][0][1] = 0.0;
      chrisD[3][3][0][2] = 0.0;
      chrisD[3][3][0][3] = 0.0;
      chrisD[3][3][1][0] = 0.0;
      chrisD[3][3][1][1] = -t47;
      chrisD[3][3][1][2] = -t49;
      chrisD[3][3][1][3] = -t51;
      chrisD[3][3][2][0] = 0.0;
      chrisD[3][3][2][1] = -t49;
      chrisD[3][3][2][2] = -t53;
      chrisD[3][3][2][3] = -t55;
      chrisD[3][3][3][0] = 0.0;
      chrisD[3][3][3][1] = t51;
      chrisD[3][3][3][2] = t55;
      chrisD[3][3][3][3] = t57;
	  
  return true;      
}

/*! Transform local 4-direction to coordinate 4-direction.
 *
 *  \param  pos  :  pointer to position array.
 *  \param  ldir :  pointer to local direction array.
 *  \param  dir  :  pointer to calculated coordinate direction array.
 *  \param  type :  type of tetrad.
 */
void
MetricExtremeReissnerNordstromDihole :: localToCoord ( const double* pos, const double* ldir, double* dir, 
                                                       enum_nat_tetrad_type )
{ 
  double U = calc_U(pos);
  
  dir[0] = ldir[0]*U;
  dir[1] = ldir[1]/U;
  dir[2] = ldir[2]/U;
  dir[3] = ldir[3]/U;
}

/*! Transform coordinate 4-direction to local 4-direction.
 *
 *  \param  pos  :  pointer to position array.
 *  \param  cdir :  pointer to coordinate direction.
 *  \param  ldir :  pointer to calculated local direction array.
 *  \param  type :  type of tetrad.
 */
void 
MetricExtremeReissnerNordstromDihole :: coordToLocal ( const double* pos, const double* cdir, double* ldir, 
                                                       enum_nat_tetrad_type )
{
  double U = calc_U(pos);
  
  ldir[0] = cdir[0]/U;
  ldir[1] = cdir[1]*U;
  ldir[2] = cdir[2]*U;
  ldir[3] = cdir[3]*U;
}


/*! Test break condition.
 *
 *  \param pos    : pointer to position array.
 *  \return true  : radial position r < 0.0 or  r^2<=(1.0+eps)*rs^2.
 *  \return false : position is valid.
 */
bool 
MetricExtremeReissnerNordstromDihole :: breakCondition ( const double* )
{
  bool br = false;

  return br;
}

/*! Tests whether the constraint equation is fulfilled.
 *
 *  The constraint equation for lightlike and timelike geodesics reads:  
 \verbatim
     sum = g_{\mu\nu} dot(x)^{\mu} dot(x)^{\nu} - kappa c^2 = 0.
 \endverbatim
 *  \param  y[]   : pointer to position and direction coordinates.
 *  \param  kappa : timelike (-1.0), lightlike (0.0).
 *  \return double : sum.
 */
double 
MetricExtremeReissnerNordstromDihole :: testConstraint ( const double y[], const double kappa )
{
  double U = calc_U(y);
  
  double dt = y[4];
  double dx = y[5];
  double dy = y[6];
  double dz = y[7];

  double sum = -kappa;
  sum += -dt*dt/(U*U) + U*U*(dx*dx + dy*dy + dz*dz);
  return sum;
}


/*! Set parameter 'pName' to 'val'.
 *
 *  Set 'mass' or 'lambda' parameter.
 */
bool 
MetricExtremeReissnerNordstromDihole :: setParam ( std::string pName, double val )
{
  Metric::setParam(pName,val);

  if (pName=="m1")
    mM1 = val;
  else if (pName=="m2")
    mM2 = val;
  return true;
}

/*! Transform point p to 2+1 coordinates.
 *
 *  \param  p  : point in proper metric coordinates.
 *  \param  cp : reference to transformed point.
 *  \return true : success.
 */
bool
MetricExtremeReissnerNordstromDihole :: transToTwoPlusOne  ( vec4 p, vec4 &cp )
{
  double x = p[1];
  double y = p[2];
  double z = p[3];
  double rho = sqrt(x*x+y*y);

  cp = vec4( p[0], rho, z, p[0] );
  return true;
}

/*! Generate report.
 */
bool    
MetricExtremeReissnerNordstromDihole :: report ( const vec4 , const vec4 , std::string &text )
{
  std::stringstream ss;
  ss << "Report for extreme ReissnerNordstrom Dihole metric\n\tcoordinate : (t,x,y,z)\n";
  ss << "---------------------------------------------------------\n";
  ss << "  physical units ................................. no\n";
  
  text = ss.str();
  return true;
}


// ********************************* protected methods *****************************
/*!
 */
void 
MetricExtremeReissnerNordstromDihole :: setStandardValues( )
{
  mInitPos[0] = 0.0; 
  mInitPos[1] = 10.0;
  mInitPos[2] = 0.0;
  mInitPos[3] = 0.0;
  mInitDir[0] = 1.0;
  mInitDir[1] = 0.0;
  mInitDir[2] = 0.0;
   
  mCoordNames[0] = std::string("t");
  mCoordNames[1] = std::string("x");
  mCoordNames[2] = std::string("y");
  mCoordNames[3] = std::string("z"); 
}

void 
MetricExtremeReissnerNordstromDihole :: calc_r ( const double y[] )
{
  r1 = sqrt(y[1]*y[1] + y[2]*y[2] + (y[3]-1.0)*(y[3]-1.0));
  r2 = sqrt(y[1]*y[1] + y[2]*y[2] + (y[3]+1.0)*(y[3]+1.0));
}

double 
MetricExtremeReissnerNordstromDihole :: calc_U  ( const double y[] )
{  
  calc_r(y);
  return 1.0 + mM1/r1 + mM2/r2;
}

void 
MetricExtremeReissnerNordstromDihole :: calc_dU ( const double y[], double &Ux, double &Uy, double &Uz )
{
  double xx = y[1];
  double yy = y[2];
  double zz = y[3];
  
  double r1hm3 = pow(r1,-3.0);
  double r2hm3 = pow(r2,-3.0);
  
  Ux = -mM1*xx*r1hm3 - mM2*xx*r2hm3;
  Uy = -mM1*yy*r1hm3 - mM2*yy*r2hm3;  
  Uz = -mM1*(zz-1.0)*r1hm3 - mM2*(zz+1.0)*r2hm3;
}

void 
MetricExtremeReissnerNordstromDihole :: calc_ddU ( const double y[], double &Uxx, double &Uyy, double &Uzz,
                                                   double &Uxy, double &Uxz, double &Uyz )
{
  double xx = y[1];
  double yy = y[2];
  double zz = y[3];
  
  double r1hm3 = pow(r1,-3.0);
  double r2hm3 = pow(r2,-3.0);
  double r1hm5 = pow(r1,-5.0);
  double r2hm5 = pow(r2,-5.0);	
  
  Uxx = 3.0*mM1*xx*xx*r1hm5 - mM1*r1hm3 + 3.0*mM2*xx*xx*r2hm5 - mM2*r2hm3;
  Uyy = 3.0*mM1*yy*yy*r1hm5 - mM1*r1hm3 + 3.0*mM2*yy*yy*r2hm5 - mM2*r2hm3;
  Uzz = 3.0*mM1*(zz-1.0)*(zz-1.0)*r1hm5 - mM1*r1hm3 + 3.0*mM2*(zz+1.0)*(zz+1.0)*r2hm5 - mM2*r2hm3;
  
  Uxy = 3.0*mM1*xx*yy*r1hm5 + 3.0*mM2*xx*yy*r2hm5;
  Uxz = 3.0*mM1*(zz-1.0)*xx*r1hm5 + 3.0*mM2*(zz+1.0)*xx*r2hm5;
  Uyz = 3.0*mM1*(zz-1.0)*yy*r1hm5 + 3.0*mM2*(zz+1.0)*yy*r2hm5;
}
													   
} // end namespace m4d
