// -------------------------------------------------------------------------------
/*
   m4dMetricPTD_AI.cpp

  Copyright (c) 2010  Thomas Mueller, Frank Grave, Felix Beslmeisl


   This file is part of the m4d-library.

   The m4d-library is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   The m4d-library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the m4d-library.  If not, see <http://www.gnu.org/licenses/>.

*/
// -------------------------------------------------------------------------------

#include "m4dMetricPTD_AI.h"

namespace m4d
{

#define eps 1.0e-6


/*! Standard constructor for the metric.
 *
 * \param  b : proportional to the mass of the black hole.
 */
MetricPTD_AI :: MetricPTD_AI ( double b)
{
  mMetricName  = "Petrov_Type_D_AI_ES";
  setCoordType(enum_coordinate_spherical);

  mPhysicalUnits = enum_physical_constants_geom;
  mSpeedOfLight = 1.0;
  mGravConstant = 1.0;

  mDrawTypes.push_back(enum_draw_effpoti);

  Par_b = b;

  addParam("b",Par_b);

  setStandardValues();

//  mLocTeds.push_back(enum_nat_tetrad_static);

}

/*! Standard destructor for the metric.
 *
 */
MetricPTD_AI :: ~MetricPTD_AI()
{

}


// *********************************** public methods ******************************
/*! Calculate the contravariant metric components at position 'pos'.
 *
 *  \param pos : pointer to position.
 */
bool
MetricPTD_AI :: calculateMetric  ( const double* pos )
{
  double r     = pos[1];
  double theta = pos[2];
  double b = Par_b;

      g_compts[0][0] = -1+b/r;
      g_compts[0][1] = 0.0;
      g_compts[0][2] = 0.0;
      g_compts[0][3] = 0.0;

      g_compts[1][1] = 1/(1-b/r);
      g_compts[1][2] = 0.0;
      g_compts[1][3] = 0.0;

      g_compts[2][2] = r*r;
      g_compts[2][3] = 0.0;

      g_compts[3][3] = (r*sin(theta))*(r*sin(theta));

     //Symmetry determines left components.
      g_compts[1][0] = g_compts[0][1];

      g_compts[2][0] = g_compts[0][2];
      g_compts[2][1] = g_compts[1][2];

      g_compts[3][0] = g_compts[0][3];
      g_compts[3][1] = g_compts[1][3];
      g_compts[3][2] = g_compts[2][3];

  return true;
}

/*! Calculate the Christoffel symbols of the second kind at position 'pos'.
 *
 *  \param pos : pointer to position.
 */
bool
MetricPTD_AI :: calculateChristoffels  ( const double* pos )
{
  double r     = pos[1];
  double theta = pos[2];
  double b = Par_b;

  double      t1 = -r+b;
  double      t2 = r*r;
  double      t8 = 1/r;
  double      t12 = t8/t1*b/2.0;
  double      t13 = sin(theta);
  double      t15 = cos(theta);
  double      t16 = 1/t13*t15;
  double      t17 = t13*t13;
        christoffel[0][0][0] = 0.0;
        christoffel[0][0][1] = -t1/t2/r*b/2.0;
        christoffel[0][0][2] = 0.0;
        christoffel[0][0][3] = 0.0;
        christoffel[0][1][0] = -t12;
        christoffel[0][1][1] = 0.0;
        christoffel[0][1][2] = 0.0;
        christoffel[0][1][3] = 0.0;
        christoffel[0][2][0] = 0.0;
        christoffel[0][2][1] = 0.0;
        christoffel[0][2][2] = 0.0;
        christoffel[0][2][3] = 0.0;
        christoffel[0][3][0] = 0.0;
        christoffel[0][3][1] = 0.0;
        christoffel[0][3][2] = 0.0;
        christoffel[0][3][3] = 0.0;
        christoffel[1][0][0] = -t12;
        christoffel[1][0][1] = 0.0;
        christoffel[1][0][2] = 0.0;
        christoffel[1][0][3] = 0.0;
        christoffel[1][1][0] = 0.0;
        christoffel[1][1][1] = t12;
        christoffel[1][1][2] = 0.0;
        christoffel[1][1][3] = 0.0;
        christoffel[1][2][0] = 0.0;
        christoffel[1][2][1] = 0.0;
        christoffel[1][2][2] = t8;
        christoffel[1][2][3] = 0.0;
        christoffel[1][3][0] = 0.0;
        christoffel[1][3][1] = 0.0;
        christoffel[1][3][2] = 0.0;
        christoffel[1][3][3] = t8;
        christoffel[2][0][0] = 0.0;
        christoffel[2][0][1] = 0.0;
        christoffel[2][0][2] = 0.0;
        christoffel[2][0][3] = 0.0;
        christoffel[2][1][0] = 0.0;
        christoffel[2][1][1] = 0.0;
        christoffel[2][1][2] = t8;
        christoffel[2][1][3] = 0.0;
        christoffel[2][2][0] = 0.0;
        christoffel[2][2][1] = t1;
        christoffel[2][2][2] = 0.0;
        christoffel[2][2][3] = 0.0;
        christoffel[2][3][0] = 0.0;
        christoffel[2][3][1] = 0.0;
        christoffel[2][3][2] = 0.0;
        christoffel[2][3][3] = t16;
        christoffel[3][0][0] = 0.0;
        christoffel[3][0][1] = 0.0;
        christoffel[3][0][2] = 0.0;
        christoffel[3][0][3] = 0.0;
        christoffel[3][1][0] = 0.0;
        christoffel[3][1][1] = 0.0;
        christoffel[3][1][2] = 0.0;
        christoffel[3][1][3] = t8;
        christoffel[3][2][0] = 0.0;
        christoffel[3][2][1] = 0.0;
        christoffel[3][2][2] = 0.0;
        christoffel[3][2][3] = t16;
        christoffel[3][3][0] = 0.0;
        christoffel[3][3][1] = t1*t17;
        christoffel[3][3][2] = -t13*t15;
        christoffel[3][3][3] = 0.0;


  return true;
}

/*! Calculate Jacobi matrix.
 *
 *  \param pos : pointer to position.
 */
bool
  MetricPTD_AI :: calculateChrisD ( const double* pos )
{
    double r     = pos[1];
    double theta = pos[2];
    double b = Par_b;

         double t1 = 2.0*r;
         double  t5 = r*r;
         double  t6 = t5*t5;
         double  t12 = -r+b;
         double  t13 = t12*t12;
         double  t15 = 1/t5;
         double  t18 = b*(-t1+b)/t13*t15/2.0;
         double  t19 = cos(theta);
         double  t20 = t19*t19;
         double  t21 = sin(theta);
         double  t22 = t21*t21;
         double  t25 = (t20+t22)/t22;
          chrisD[0][0][0][0] = 0.0;
          chrisD[0][0][0][1] = 0.0;
          chrisD[0][0][0][2] = 0.0;
          chrisD[0][0][0][3] = 0.0;
          chrisD[0][0][1][0] = 0.0;
          chrisD[0][0][1][1] = b*(-t1+3.0*b)/t6/2.0;
          chrisD[0][0][1][2] = 0.0;
          chrisD[0][0][1][3] = 0.0;
          chrisD[0][0][2][0] = 0.0;
          chrisD[0][0][2][1] = 0.0;
          chrisD[0][0][2][2] = 0.0;
          chrisD[0][0][2][3] = 0.0;
          chrisD[0][0][3][0] = 0.0;
          chrisD[0][0][3][1] = 0.0;
          chrisD[0][0][3][2] = 0.0;
          chrisD[0][0][3][3] = 0.0;
          chrisD[0][1][0][0] = 0.0;
          chrisD[0][1][0][1] = t18;
          chrisD[0][1][0][2] = 0.0;
          chrisD[0][1][0][3] = 0.0;
          chrisD[0][1][1][0] = 0.0;
          chrisD[0][1][1][1] = 0.0;
          chrisD[0][1][1][2] = 0.0;
          chrisD[0][1][1][3] = 0.0;
          chrisD[0][1][2][0] = 0.0;
          chrisD[0][1][2][1] = 0.0;
          chrisD[0][1][2][2] = 0.0;
          chrisD[0][1][2][3] = 0.0;
          chrisD[0][1][3][0] = 0.0;
          chrisD[0][1][3][1] = 0.0;
          chrisD[0][1][3][2] = 0.0;
          chrisD[0][1][3][3] = 0.0;
          chrisD[0][2][0][0] = 0.0;
          chrisD[0][2][0][1] = 0.0;
          chrisD[0][2][0][2] = 0.0;
          chrisD[0][2][0][3] = 0.0;
          chrisD[0][2][1][0] = 0.0;
          chrisD[0][2][1][1] = 0.0;
          chrisD[0][2][1][2] = 0.0;
          chrisD[0][2][1][3] = 0.0;
          chrisD[0][2][2][0] = 0.0;
          chrisD[0][2][2][1] = 0.0;
          chrisD[0][2][2][2] = 0.0;
          chrisD[0][2][2][3] = 0.0;
          chrisD[0][2][3][0] = 0.0;
          chrisD[0][2][3][1] = 0.0;
          chrisD[0][2][3][2] = 0.0;
          chrisD[0][2][3][3] = 0.0;
          chrisD[0][3][0][0] = 0.0;
          chrisD[0][3][0][1] = 0.0;
          chrisD[0][3][0][2] = 0.0;
          chrisD[0][3][0][3] = 0.0;
          chrisD[0][3][1][0] = 0.0;
          chrisD[0][3][1][1] = 0.0;
          chrisD[0][3][1][2] = 0.0;
          chrisD[0][3][1][3] = 0.0;
          chrisD[0][3][2][0] = 0.0;
          chrisD[0][3][2][1] = 0.0;
          chrisD[0][3][2][2] = 0.0;
          chrisD[0][3][2][3] = 0.0;
          chrisD[0][3][3][0] = 0.0;
          chrisD[0][3][3][1] = 0.0;
          chrisD[0][3][3][2] = 0.0;
          chrisD[0][3][3][3] = 0.0;
          chrisD[1][0][0][0] = 0.0;
          chrisD[1][0][0][1] = t18;
          chrisD[1][0][0][2] = 0.0;
          chrisD[1][0][0][3] = 0.0;
          chrisD[1][0][1][0] = 0.0;
          chrisD[1][0][1][1] = 0.0;
          chrisD[1][0][1][2] = 0.0;
          chrisD[1][0][1][3] = 0.0;
          chrisD[1][0][2][0] = 0.0;
          chrisD[1][0][2][1] = 0.0;
          chrisD[1][0][2][2] = 0.0;
          chrisD[1][0][2][3] = 0.0;
          chrisD[1][0][3][0] = 0.0;
          chrisD[1][0][3][1] = 0.0;
          chrisD[1][0][3][2] = 0.0;
          chrisD[1][0][3][3] = 0.0;
          chrisD[1][1][0][0] = 0.0;
          chrisD[1][1][0][1] = 0.0;
          chrisD[1][1][0][2] = 0.0;
          chrisD[1][1][0][3] = 0.0;
          chrisD[1][1][1][0] = 0.0;
          chrisD[1][1][1][1] = -t18;
          chrisD[1][1][1][2] = 0.0;
          chrisD[1][1][1][3] = 0.0;
          chrisD[1][1][2][0] = 0.0;
          chrisD[1][1][2][1] = 0.0;
          chrisD[1][1][2][2] = 0.0;
          chrisD[1][1][2][3] = 0.0;
          chrisD[1][1][3][0] = 0.0;
          chrisD[1][1][3][1] = 0.0;
          chrisD[1][1][3][2] = 0.0;
          chrisD[1][1][3][3] = 0.0;
          chrisD[1][2][0][0] = 0.0;
          chrisD[1][2][0][1] = 0.0;
          chrisD[1][2][0][2] = 0.0;
          chrisD[1][2][0][3] = 0.0;
          chrisD[1][2][1][0] = 0.0;
          chrisD[1][2][1][1] = 0.0;
          chrisD[1][2][1][2] = 0.0;
          chrisD[1][2][1][3] = 0.0;
          chrisD[1][2][2][0] = 0.0;
          chrisD[1][2][2][1] = -t15;
          chrisD[1][2][2][2] = 0.0;
          chrisD[1][2][2][3] = 0.0;
          chrisD[1][2][3][0] = 0.0;
          chrisD[1][2][3][1] = 0.0;
          chrisD[1][2][3][2] = 0.0;
          chrisD[1][2][3][3] = 0.0;
          chrisD[1][3][0][0] = 0.0;
          chrisD[1][3][0][1] = 0.0;
          chrisD[1][3][0][2] = 0.0;
          chrisD[1][3][0][3] = 0.0;
          chrisD[1][3][1][0] = 0.0;
          chrisD[1][3][1][1] = 0.0;
          chrisD[1][3][1][2] = 0.0;
          chrisD[1][3][1][3] = 0.0;
          chrisD[1][3][2][0] = 0.0;
          chrisD[1][3][2][1] = 0.0;
          chrisD[1][3][2][2] = 0.0;
          chrisD[1][3][2][3] = 0.0;
          chrisD[1][3][3][0] = 0.0;
          chrisD[1][3][3][1] = -t15;
          chrisD[1][3][3][2] = 0.0;
          chrisD[1][3][3][3] = 0.0;
          chrisD[2][0][0][0] = 0.0;
          chrisD[2][0][0][1] = 0.0;
          chrisD[2][0][0][2] = 0.0;
          chrisD[2][0][0][3] = 0.0;
          chrisD[2][0][1][0] = 0.0;
          chrisD[2][0][1][1] = 0.0;
          chrisD[2][0][1][2] = 0.0;
          chrisD[2][0][1][3] = 0.0;
          chrisD[2][0][2][0] = 0.0;
          chrisD[2][0][2][1] = 0.0;
          chrisD[2][0][2][2] = 0.0;
          chrisD[2][0][2][3] = 0.0;
          chrisD[2][0][3][0] = 0.0;
          chrisD[2][0][3][1] = 0.0;
          chrisD[2][0][3][2] = 0.0;
          chrisD[2][0][3][3] = 0.0;
          chrisD[2][1][0][0] = 0.0;
          chrisD[2][1][0][1] = 0.0;
          chrisD[2][1][0][2] = 0.0;
          chrisD[2][1][0][3] = 0.0;
          chrisD[2][1][1][0] = 0.0;
          chrisD[2][1][1][1] = 0.0;
          chrisD[2][1][1][2] = 0.0;
          chrisD[2][1][1][3] = 0.0;
          chrisD[2][1][2][0] = 0.0;
          chrisD[2][1][2][1] = -t15;
          chrisD[2][1][2][2] = 0.0;
          chrisD[2][1][2][3] = 0.0;
          chrisD[2][1][3][0] = 0.0;
          chrisD[2][1][3][1] = 0.0;
          chrisD[2][1][3][2] = 0.0;
          chrisD[2][1][3][3] = 0.0;
          chrisD[2][2][0][0] = 0.0;
          chrisD[2][2][0][1] = 0.0;
          chrisD[2][2][0][2] = 0.0;
          chrisD[2][2][0][3] = 0.0;
          chrisD[2][2][1][0] = 0.0;
          chrisD[2][2][1][1] = -1.0;
          chrisD[2][2][1][2] = 0.0;
          chrisD[2][2][1][3] = 0.0;
          chrisD[2][2][2][0] = 0.0;
          chrisD[2][2][2][1] = 0.0;
          chrisD[2][2][2][2] = 0.0;
          chrisD[2][2][2][3] = 0.0;
          chrisD[2][2][3][0] = 0.0;
          chrisD[2][2][3][1] = 0.0;
          chrisD[2][2][3][2] = 0.0;
          chrisD[2][2][3][3] = 0.0;
          chrisD[2][3][0][0] = 0.0;
          chrisD[2][3][0][1] = 0.0;
          chrisD[2][3][0][2] = 0.0;
          chrisD[2][3][0][3] = 0.0;
          chrisD[2][3][1][0] = 0.0;
          chrisD[2][3][1][1] = 0.0;
          chrisD[2][3][1][2] = 0.0;
          chrisD[2][3][1][3] = 0.0;
          chrisD[2][3][2][0] = 0.0;
          chrisD[2][3][2][1] = 0.0;
          chrisD[2][3][2][2] = 0.0;
          chrisD[2][3][2][3] = 0.0;
          chrisD[2][3][3][0] = 0.0;
          chrisD[2][3][3][1] = 0.0;
          chrisD[2][3][3][2] = -t25;
          chrisD[2][3][3][3] = 0.0;
          chrisD[3][0][0][0] = 0.0;
          chrisD[3][0][0][1] = 0.0;
          chrisD[3][0][0][2] = 0.0;
          chrisD[3][0][0][3] = 0.0;
          chrisD[3][0][1][0] = 0.0;
          chrisD[3][0][1][1] = 0.0;
          chrisD[3][0][1][2] = 0.0;
          chrisD[3][0][1][3] = 0.0;
          chrisD[3][0][2][0] = 0.0;
          chrisD[3][0][2][1] = 0.0;
          chrisD[3][0][2][2] = 0.0;
          chrisD[3][0][2][3] = 0.0;
          chrisD[3][0][3][0] = 0.0;
          chrisD[3][0][3][1] = 0.0;
          chrisD[3][0][3][2] = 0.0;
          chrisD[3][0][3][3] = 0.0;
          chrisD[3][1][0][0] = 0.0;
          chrisD[3][1][0][1] = 0.0;
          chrisD[3][1][0][2] = 0.0;
          chrisD[3][1][0][3] = 0.0;
          chrisD[3][1][1][0] = 0.0;
          chrisD[3][1][1][1] = 0.0;
          chrisD[3][1][1][2] = 0.0;
          chrisD[3][1][1][3] = 0.0;
          chrisD[3][1][2][0] = 0.0;
          chrisD[3][1][2][1] = 0.0;
          chrisD[3][1][2][2] = 0.0;
          chrisD[3][1][2][3] = 0.0;
          chrisD[3][1][3][0] = 0.0;
          chrisD[3][1][3][1] = -t15;
          chrisD[3][1][3][2] = 0.0;
          chrisD[3][1][3][3] = 0.0;
          chrisD[3][2][0][0] = 0.0;
          chrisD[3][2][0][1] = 0.0;
          chrisD[3][2][0][2] = 0.0;
          chrisD[3][2][0][3] = 0.0;
          chrisD[3][2][1][0] = 0.0;
          chrisD[3][2][1][1] = 0.0;
          chrisD[3][2][1][2] = 0.0;
          chrisD[3][2][1][3] = 0.0;
          chrisD[3][2][2][0] = 0.0;
          chrisD[3][2][2][1] = 0.0;
          chrisD[3][2][2][2] = 0.0;
          chrisD[3][2][2][3] = 0.0;
          chrisD[3][2][3][0] = 0.0;
          chrisD[3][2][3][1] = 0.0;
          chrisD[3][2][3][2] = -t25;
          chrisD[3][2][3][3] = 0.0;
          chrisD[3][3][0][0] = 0.0;
          chrisD[3][3][0][1] = 0.0;
          chrisD[3][3][0][2] = 0.0;
          chrisD[3][3][0][3] = 0.0;
          chrisD[3][3][1][0] = 0.0;
          chrisD[3][3][1][1] = -t22;
          chrisD[3][3][1][2] = 2.0*t12*t21*t19;
          chrisD[3][3][1][3] = 0.0;
          chrisD[3][3][2][0] = 0.0;
          chrisD[3][3][2][1] = 0.0;
          chrisD[3][3][2][2] = -t20+t22;
          chrisD[3][3][2][3] = 0.0;
          chrisD[3][3][3][0] = 0.0;
          chrisD[3][3][3][1] = 0.0;
          chrisD[3][3][3][2] = 0.0;
          chrisD[3][3][3][3] = 0.0;

  return true;
}

/*! Transform local 4-direction to coordinate 4-direction.
 *
 *  \param  pos  :  pointer to position array.
 *  \param  ldir :  pointer to local direction array.
 *  \param  dir  :  pointer to calculated coordinate direction array.
 *  \param  type :  type of tetrad.
 */
void
MetricPTD_AI :: localToCoord ( const double* pos, const double* ldir, double* dir,
                               enum_nat_tetrad_type  )
{
    double r     = pos[1];
    double theta = pos[2];
    double w = sqrt(1.0-Par_b/r);

    dir[0] = ldir[0]/w;
    dir[1] = ldir[1]*w;
    dir[2] = ldir[2]/r;
    dir[3] = ldir[3]/(r*sin(theta));
}

/*! Transform coordinate 4-direction to local 4-direction.
 *
 *  \param  pos  :  pointer to position array.
 *  \param  cdir :  pointer to coordinate direction.
 *  \param  ldir :  pointer to calculated local direction array.
 *  \param  type :  type of tetrad.
 */
void
MetricPTD_AI :: coordToLocal   ( const double* pos, const double* cdir, double* ldir,
                                           enum_nat_tetrad_type   )
{
    double r     = pos[1];
    double theta = pos[2];
    double w = sqrt(1.0-Par_b/r);

    ldir[0] = cdir[0]*w;
    ldir[1] = cdir[1]/w;
    ldir[2] = cdir[2]*r;
    ldir[3] = cdir[3]*r*sin(theta);
}


/*! Tests break condition.
 *
 *  \param pos  :  position.
 *  \return true  : radial position r < 0.0 or geodesic reaches event horizont.
 *  \return false : position is valid.
 */
bool
MetricPTD_AI :: breakCondition ( const double* pos)
{
    bool br = false;

    if ((pos[1]<0.0) || (pos[1]<=(1.0+eps)*Par_b)) { br=true; }
    return br;
}

/*! Calculates the constants of Motion.
 *
 *  \param pos : initial position.
 *  \param cdir : initial four-direction.
 */
void
MetricPTD_AI :: calcConstantsOfMotion (const vec4 pos, const vec4 cdir)
{
   double sinp2 = sin(pos[2]);
   double p14   = pos[1]*pos[1]*pos[1]*pos[1];
   double b     = Par_b;
   C0 = cdir[0]*cdir[0]*(1.0 - b/pos[1])*(1.0 - b/pos[1]);
   C2 = cdir[3]*cdir[3]*p14*sinp2*sinp2*sinp2*sinp2;
   K  = cdir[2]*cdir[2]*p14 + cdir[3]*cdir[3]*p14*sinp2*sinp2;
   m0 = -K/pos[1]/pos[1]  - cdir[1]*cdir[1]*pos[1]/(pos[1] - b) + cdir[0]*cdir[0]*(pos[1] - b)/pos[1];
}

/*! Effective potential.
 *  \param pos : initial position.
 *  \param cdir : initial four-direction.
 *  \param type : geodesic type.
 *  \param x : abscissa value.
 *  \param val : reference to effective potential value.
 *  \return true : effective potential exists at x.
 */
bool
MetricPTD_AI :: effPotentialValue  ( const vec4 pos, const vec4 cdir ,enum_geodesic_type type, const double x, double &val )
{
  double kappa = 0.0;
  if (type==enum_geodesic_timelike)
    kappa = -mSign;

  if (x<=0.0)
     return false;

  double r = x;
  double b=Par_b;
  calcConstantsOfMotion(pos, cdir);

  val = (r-b)/r/r/r*K+(r-b)/r*m0-C0;
  return true;
}

/*! Total energy.
 *  \param pos : initial position.
 *  \param cdir : initial four-direction.
 *  \param x : abscissa value.
 *  \param val : reference to total energy value.
 *  \return true : effective potential exists at x.
 */
bool
MetricPTD_AI :: totEnergy ( const vec4 , const vec4 , const double , double &val )
{
  val = 0.0;
  return true;
}

/*! Set parameter 'pName' to 'val'.
 *
 *
 */
bool
MetricPTD_AI :: setParam ( std::string pName, double val )
{
  Metric::setParam(pName,val);
  if (pName=="b") Par_b = val;
  return true;
}



/*! Generate report.
 */
bool
MetricPTD_AI :: report ( const vec4 pos, const vec4 cdir, std::string &text )
{
    std::stringstream ss;
    ss << "Report for AI (Schwarzschild metric)\n\tcoordinates : (t,r,theta,phi)\n";
    ss << "---------------------------------------------------------------\n";
    ss << "Coordinate Ranges:" << std::endl;
    ss << "        r: (0 < b < r) or (b < 0 < r)"<< std::endl;
    ss << "    theta: (0 <= theta <= pi)" << std::endl;
    ss << "      phi: (phi mod 2*pi)" << std::endl;
    ss << "---------------------------------------------------------------\n";
    ss << "Parameter Range:" << std::endl;
    ss << "        b: arbitrary"<< std::endl;
    ss << "---------------------------------------------------------------\n";
    ss.precision(DEF_FIXED_REPORT_PRECISION);
    ss.setf(std::ios::fixed);

    calcConstantsOfMotion(pos, cdir);
    ss << "Constants of motion:"<< std::endl;
    ss << "  C_0^2 = " << C0 << std::endl;
    ss << "  C_2^2 = " << C2 << std::endl;
    ss << "      K = " << K  << std::endl;
    //ss << "  m_0^2 = " << m0 << std::endl;
    ss << "---------------------------------------------------------------\n";
    ss << " Further information see report of Metrik 'Schwarzschild'.";

    text = ss.str();
    return true;
}

// *************************** specific  public methods ****************************
//None
// ********************************* protected methods *****************************
/*!
 */
void
MetricPTD_AI :: setStandardValues( )
{
  mInitPos[0] = 0.0;
  mInitPos[1] = 6.0*Par_b;
  mInitPos[2] = M_PI_2;
  mInitPos[3] = 0.0;
  mInitDir[0] = 1.0;
  mInitDir[1] = 0.0;
  mInitDir[2] = 0.0;

  mCoordNames[0] = std::string("t");
  mCoordNames[1] = std::string("r");
  mCoordNames[2] = std::string("theta");
  mCoordNames[3] = std::string("phi");
}


} // end namespace m4d

