// -------------------------------------------------------------------------------
/*
   m4dMetricStraightSpinningString.cpp
 
  Copyright (c) 2009,2010  Thomas Mueller, Frank Grave
 
 
   This file is part of the m4d-library.
 
   The m4d-library is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
 
   The m4d-library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with the m4d-library.  If not, see <http://www.gnu.org/licenses/>.
 
*/
// -------------------------------------------------------------------------------

#include "m4dMetricStraightSpinningString.h"

namespace m4d
{

#define eps 1.0e-6


/*! Standard constructor for the Kottler metric.
 *
 * \param  a : parameter.
 * \param  k : parameter.
 */
MetricStraightSpinningString :: MetricStraightSpinningString ( double a, double k )
{
  mMetricName  = "StraightSpinningString";
  setCoordType(enum_coordinate_cylinder);
    
  mPhysicalUnits = enum_physical_constants_geom;
  mSpeedOfLight = 1.0;
  mGravConstant = 1.0;
  
  addParam("a");
  setParam("a",a);
  mA = a;

  addParam("k");
  setParam("k",k);
  mK = k;
  
  mLocTeds.push_back(enum_nat_tetrad_static);
  mLocTeds.push_back(enum_nat_tetrad_lnrf);

  mDrawTypes.push_back(enum_draw_effpoti);

  setStandardValues();
}

MetricStraightSpinningString :: ~MetricStraightSpinningString()
{
}


// *********************************** public methods ******************************
/*! Calculate the contravariant metric components at position 'pos'.
 *
 *  \param pos : pointer to position.
 */
bool 
MetricStraightSpinningString :: calculateMetric  ( const double* pos )
{
  double rho = pos[1];

  double a = mA;
  double k = mK;
  double c = mSpeedOfLight;

  double t1 = c*c;
  double t2 = c*a;
  double t3 = a*a;
  double t4 = k*k;
  double t5 = rho*rho;

  g_compts[0][0] = -t1;
  g_compts[0][1] = 0.0;
  g_compts[0][2] = t2;
  g_compts[0][3] = 0.0;
  g_compts[1][0] = 0.0;
  g_compts[1][1] = 1.0;
  g_compts[1][2] = 0.0;
  g_compts[1][3] = 0.0;
  g_compts[2][0] = t2;
  g_compts[2][1] = 0.0;
  g_compts[2][2] = -t3+t4*t5;
  g_compts[2][3] = 0.0;
  g_compts[3][0] = 0.0;
  g_compts[3][1] = 0.0;
  g_compts[3][2] = 0.0;
  g_compts[3][3] = 1.0;

  return true;
}

/*! Calculate the Christoffel symbols of the second kind at position 'pos'.
 *
 *  \param pos : pointer to position.
 */
bool 
MetricStraightSpinningString :: calculateChristoffels  ( const double* pos )
{
  double rho = pos[1];

  double a = mA;
  double k = mK;
  double c = mSpeedOfLight;

  double t3 = 1/rho;
  double t4 = 1/c*a*t3;
  double t5 = k*k;

  christoffel[0][0][0] = 0.0;
  christoffel[0][0][1] = 0.0;
  christoffel[0][0][2] = 0.0;
  christoffel[0][0][3] = 0.0;
  christoffel[0][1][0] = 0.0;
  christoffel[0][1][1] = 0.0;
  christoffel[0][1][2] = 0.0;
  christoffel[0][1][3] = 0.0;
  christoffel[0][2][0] = 0.0;
  christoffel[0][2][1] = 0.0;
  christoffel[0][2][2] = 0.0;
  christoffel[0][2][3] = 0.0;
  christoffel[0][3][0] = 0.0;
  christoffel[0][3][1] = 0.0;
  christoffel[0][3][2] = 0.0;
  christoffel[0][3][3] = 0.0;
  christoffel[1][0][0] = 0.0;
  christoffel[1][0][1] = 0.0;
  christoffel[1][0][2] = 0.0;
  christoffel[1][0][3] = 0.0;
  christoffel[1][1][0] = 0.0;
  christoffel[1][1][1] = 0.0;
  christoffel[1][1][2] = 0.0;
  christoffel[1][1][3] = 0.0;
  christoffel[1][2][0] = t4;
  christoffel[1][2][1] = 0.0;
  christoffel[1][2][2] = t3;
  christoffel[1][2][3] = 0.0;
  christoffel[1][3][0] = 0.0;
  christoffel[1][3][1] = 0.0;
  christoffel[1][3][2] = 0.0;
  christoffel[1][3][3] = 0.0;
  christoffel[2][0][0] = 0.0;
  christoffel[2][0][1] = 0.0;
  christoffel[2][0][2] = 0.0;
  christoffel[2][0][3] = 0.0;
  christoffel[2][1][0] = t4;
  christoffel[2][1][1] = 0.0;
  christoffel[2][1][2] = t3;
  christoffel[2][1][3] = 0.0;
  christoffel[2][2][0] = 0.0;
  christoffel[2][2][1] = -t5*rho;
  christoffel[2][2][2] = 0.0;
  christoffel[2][2][3] = 0.0;
  christoffel[2][3][0] = 0.0;
  christoffel[2][3][1] = 0.0;
  christoffel[2][3][2] = 0.0;
  christoffel[2][3][3] = 0.0;
  christoffel[3][0][0] = 0.0;
  christoffel[3][0][1] = 0.0;
  christoffel[3][0][2] = 0.0;
  christoffel[3][0][3] = 0.0;
  christoffel[3][1][0] = 0.0;
  christoffel[3][1][1] = 0.0;
  christoffel[3][1][2] = 0.0;
  christoffel[3][1][3] = 0.0;
  christoffel[3][2][0] = 0.0;
  christoffel[3][2][1] = 0.0;
  christoffel[3][2][2] = 0.0;
  christoffel[3][2][3] = 0.0;
  christoffel[3][3][0] = 0.0;
  christoffel[3][3][1] = 0.0;
  christoffel[3][3][2] = 0.0;
  christoffel[3][3][3] = 0.0;

  return true;
}

/*! Calculate Jacobi matrix. 
 *
 *  \param pos : pointer to position.
 */
bool 
MetricStraightSpinningString :: calculateChrisD ( const double* pos )
{
  double rho = pos[1];

  double a = mA;
  double k = mK;
  double c = mSpeedOfLight;

  double t3 = rho*rho;
  double t4 = 1/t3;
  double t5 = 1/c*a*t4;
  double t6 = k*k;

  chrisD[0][0][0][0] = 0.0;
  chrisD[0][0][0][1] = 0.0;
  chrisD[0][0][0][2] = 0.0;
  chrisD[0][0][0][3] = 0.0;
  chrisD[0][0][1][0] = 0.0;
  chrisD[0][0][1][1] = 0.0;
  chrisD[0][0][1][2] = 0.0;
  chrisD[0][0][1][3] = 0.0;
  chrisD[0][0][2][0] = 0.0;
  chrisD[0][0][2][1] = 0.0;
  chrisD[0][0][2][2] = 0.0;
  chrisD[0][0][2][3] = 0.0;
  chrisD[0][0][3][0] = 0.0;
  chrisD[0][0][3][1] = 0.0;
  chrisD[0][0][3][2] = 0.0;
  chrisD[0][0][3][3] = 0.0;
  chrisD[0][1][0][0] = 0.0;
  chrisD[0][1][0][1] = 0.0;
  chrisD[0][1][0][2] = 0.0;
  chrisD[0][1][0][3] = 0.0;
  chrisD[0][1][1][0] = 0.0;
  chrisD[0][1][1][1] = 0.0;
  chrisD[0][1][1][2] = 0.0;
  chrisD[0][1][1][3] = 0.0;
  chrisD[0][1][2][0] = 0.0;
  chrisD[0][1][2][1] = 0.0;
  chrisD[0][1][2][2] = 0.0;
  chrisD[0][1][2][3] = 0.0;
  chrisD[0][1][3][0] = 0.0;
  chrisD[0][1][3][1] = 0.0;
  chrisD[0][1][3][2] = 0.0;
  chrisD[0][1][3][3] = 0.0;
  chrisD[0][2][0][0] = 0.0;
  chrisD[0][2][0][1] = 0.0;
  chrisD[0][2][0][2] = 0.0;
  chrisD[0][2][0][3] = 0.0;
  chrisD[0][2][1][0] = 0.0;
  chrisD[0][2][1][1] = 0.0;
  chrisD[0][2][1][2] = 0.0;
  chrisD[0][2][1][3] = 0.0;
  chrisD[0][2][2][0] = 0.0;
  chrisD[0][2][2][1] = 0.0;
  chrisD[0][2][2][2] = 0.0;
  chrisD[0][2][2][3] = 0.0;
  chrisD[0][2][3][0] = 0.0;
  chrisD[0][2][3][1] = 0.0;
  chrisD[0][2][3][2] = 0.0;
  chrisD[0][2][3][3] = 0.0;
  chrisD[0][3][0][0] = 0.0;
  chrisD[0][3][0][1] = 0.0;
  chrisD[0][3][0][2] = 0.0;
  chrisD[0][3][0][3] = 0.0;
  chrisD[0][3][1][0] = 0.0;
  chrisD[0][3][1][1] = 0.0;
  chrisD[0][3][1][2] = 0.0;
  chrisD[0][3][1][3] = 0.0;
  chrisD[0][3][2][0] = 0.0;
  chrisD[0][3][2][1] = 0.0;
  chrisD[0][3][2][2] = 0.0;
  chrisD[0][3][2][3] = 0.0;
  chrisD[0][3][3][0] = 0.0;
  chrisD[0][3][3][1] = 0.0;
  chrisD[0][3][3][2] = 0.0;
  chrisD[0][3][3][3] = 0.0;
  chrisD[1][0][0][0] = 0.0;
  chrisD[1][0][0][1] = 0.0;
  chrisD[1][0][0][2] = 0.0;
  chrisD[1][0][0][3] = 0.0;
  chrisD[1][0][1][0] = 0.0;
  chrisD[1][0][1][1] = 0.0;
  chrisD[1][0][1][2] = 0.0;
  chrisD[1][0][1][3] = 0.0;
  chrisD[1][0][2][0] = 0.0;
  chrisD[1][0][2][1] = 0.0;
  chrisD[1][0][2][2] = 0.0;
  chrisD[1][0][2][3] = 0.0;
  chrisD[1][0][3][0] = 0.0;
  chrisD[1][0][3][1] = 0.0;
  chrisD[1][0][3][2] = 0.0;
  chrisD[1][0][3][3] = 0.0;
  chrisD[1][1][0][0] = 0.0;
  chrisD[1][1][0][1] = 0.0;
  chrisD[1][1][0][2] = 0.0;
  chrisD[1][1][0][3] = 0.0;
  chrisD[1][1][1][0] = 0.0;
  chrisD[1][1][1][1] = 0.0;
  chrisD[1][1][1][2] = 0.0;
  chrisD[1][1][1][3] = 0.0;
  chrisD[1][1][2][0] = 0.0;
  chrisD[1][1][2][1] = 0.0;
  chrisD[1][1][2][2] = 0.0;
  chrisD[1][1][2][3] = 0.0;
  chrisD[1][1][3][0] = 0.0;
  chrisD[1][1][3][1] = 0.0;
  chrisD[1][1][3][2] = 0.0;
  chrisD[1][1][3][3] = 0.0;
  chrisD[1][2][0][0] = 0.0;
  chrisD[1][2][0][1] = -t5;
  chrisD[1][2][0][2] = 0.0;
  chrisD[1][2][0][3] = 0.0;
  chrisD[1][2][1][0] = 0.0;
  chrisD[1][2][1][1] = 0.0;
  chrisD[1][2][1][2] = 0.0;
  chrisD[1][2][1][3] = 0.0;
  chrisD[1][2][2][0] = 0.0;
  chrisD[1][2][2][1] = -t4;
  chrisD[1][2][2][2] = 0.0;
  chrisD[1][2][2][3] = 0.0;
  chrisD[1][2][3][0] = 0.0;
  chrisD[1][2][3][1] = 0.0;
  chrisD[1][2][3][2] = 0.0;
  chrisD[1][2][3][3] = 0.0;
  chrisD[1][3][0][0] = 0.0;
  chrisD[1][3][0][1] = 0.0;
  chrisD[1][3][0][2] = 0.0;
  chrisD[1][3][0][3] = 0.0;
  chrisD[1][3][1][0] = 0.0;
  chrisD[1][3][1][1] = 0.0;
  chrisD[1][3][1][2] = 0.0;
  chrisD[1][3][1][3] = 0.0;
  chrisD[1][3][2][0] = 0.0;
  chrisD[1][3][2][1] = 0.0;
  chrisD[1][3][2][2] = 0.0;
  chrisD[1][3][2][3] = 0.0;
  chrisD[1][3][3][0] = 0.0;
  chrisD[1][3][3][1] = 0.0;
  chrisD[1][3][3][2] = 0.0;
  chrisD[1][3][3][3] = 0.0;
  chrisD[2][0][0][0] = 0.0;
  chrisD[2][0][0][1] = 0.0;
  chrisD[2][0][0][2] = 0.0;
  chrisD[2][0][0][3] = 0.0;
  chrisD[2][0][1][0] = 0.0;
  chrisD[2][0][1][1] = 0.0;
  chrisD[2][0][1][2] = 0.0;
  chrisD[2][0][1][3] = 0.0;
  chrisD[2][0][2][0] = 0.0;
  chrisD[2][0][2][1] = 0.0;
  chrisD[2][0][2][2] = 0.0;
  chrisD[2][0][2][3] = 0.0;
  chrisD[2][0][3][0] = 0.0;
  chrisD[2][0][3][1] = 0.0;
  chrisD[2][0][3][2] = 0.0;
  chrisD[2][0][3][3] = 0.0;
  chrisD[2][1][0][0] = 0.0;
  chrisD[2][1][0][1] = -t5;
  chrisD[2][1][0][2] = 0.0;
  chrisD[2][1][0][3] = 0.0;
  chrisD[2][1][1][0] = 0.0;
  chrisD[2][1][1][1] = 0.0;
  chrisD[2][1][1][2] = 0.0;
  chrisD[2][1][1][3] = 0.0;
  chrisD[2][1][2][0] = 0.0;
  chrisD[2][1][2][1] = -t4;
  chrisD[2][1][2][2] = 0.0;
  chrisD[2][1][2][3] = 0.0;
  chrisD[2][1][3][0] = 0.0;
  chrisD[2][1][3][1] = 0.0;
  chrisD[2][1][3][2] = 0.0;
  chrisD[2][1][3][3] = 0.0;
  chrisD[2][2][0][0] = 0.0;
  chrisD[2][2][0][1] = 0.0;
  chrisD[2][2][0][2] = 0.0;
  chrisD[2][2][0][3] = 0.0;
  chrisD[2][2][1][0] = 0.0;
  chrisD[2][2][1][1] = -t6;
  chrisD[2][2][1][2] = 0.0;
  chrisD[2][2][1][3] = 0.0;
  chrisD[2][2][2][0] = 0.0;
  chrisD[2][2][2][1] = 0.0;
  chrisD[2][2][2][2] = 0.0;
  chrisD[2][2][2][3] = 0.0;
  chrisD[2][2][3][0] = 0.0;
  chrisD[2][2][3][1] = 0.0;
  chrisD[2][2][3][2] = 0.0;
  chrisD[2][2][3][3] = 0.0;
  chrisD[2][3][0][0] = 0.0;
  chrisD[2][3][0][1] = 0.0;
  chrisD[2][3][0][2] = 0.0;
  chrisD[2][3][0][3] = 0.0;
  chrisD[2][3][1][0] = 0.0;
  chrisD[2][3][1][1] = 0.0;
  chrisD[2][3][1][2] = 0.0;
  chrisD[2][3][1][3] = 0.0;
  chrisD[2][3][2][0] = 0.0;
  chrisD[2][3][2][1] = 0.0;
  chrisD[2][3][2][2] = 0.0;
  chrisD[2][3][2][3] = 0.0;
  chrisD[2][3][3][0] = 0.0;
  chrisD[2][3][3][1] = 0.0;
  chrisD[2][3][3][2] = 0.0;
  chrisD[2][3][3][3] = 0.0;
  chrisD[3][0][0][0] = 0.0;
  chrisD[3][0][0][1] = 0.0;
  chrisD[3][0][0][2] = 0.0;
  chrisD[3][0][0][3] = 0.0;
  chrisD[3][0][1][0] = 0.0;
  chrisD[3][0][1][1] = 0.0;
  chrisD[3][0][1][2] = 0.0;
  chrisD[3][0][1][3] = 0.0;
  chrisD[3][0][2][0] = 0.0;
  chrisD[3][0][2][1] = 0.0;
  chrisD[3][0][2][2] = 0.0;
  chrisD[3][0][2][3] = 0.0;
  chrisD[3][0][3][0] = 0.0;
  chrisD[3][0][3][1] = 0.0;
  chrisD[3][0][3][2] = 0.0;
  chrisD[3][0][3][3] = 0.0;
  chrisD[3][1][0][0] = 0.0;
  chrisD[3][1][0][1] = 0.0;
  chrisD[3][1][0][2] = 0.0;
  chrisD[3][1][0][3] = 0.0;
  chrisD[3][1][1][0] = 0.0;
  chrisD[3][1][1][1] = 0.0;
  chrisD[3][1][1][2] = 0.0;
  chrisD[3][1][1][3] = 0.0;
  chrisD[3][1][2][0] = 0.0;
  chrisD[3][1][2][1] = 0.0;
  chrisD[3][1][2][2] = 0.0;
  chrisD[3][1][2][3] = 0.0;
  chrisD[3][1][3][0] = 0.0;
  chrisD[3][1][3][1] = 0.0;
  chrisD[3][1][3][2] = 0.0;
  chrisD[3][1][3][3] = 0.0;
  chrisD[3][2][0][0] = 0.0;
  chrisD[3][2][0][1] = 0.0;
  chrisD[3][2][0][2] = 0.0;
  chrisD[3][2][0][3] = 0.0;
  chrisD[3][2][1][0] = 0.0;
  chrisD[3][2][1][1] = 0.0;
  chrisD[3][2][1][2] = 0.0;
  chrisD[3][2][1][3] = 0.0;
  chrisD[3][2][2][0] = 0.0;
  chrisD[3][2][2][1] = 0.0;
  chrisD[3][2][2][2] = 0.0;
  chrisD[3][2][2][3] = 0.0;
  chrisD[3][2][3][0] = 0.0;
  chrisD[3][2][3][1] = 0.0;
  chrisD[3][2][3][2] = 0.0;
  chrisD[3][2][3][3] = 0.0;
  chrisD[3][3][0][0] = 0.0;
  chrisD[3][3][0][1] = 0.0;
  chrisD[3][3][0][2] = 0.0;
  chrisD[3][3][0][3] = 0.0;
  chrisD[3][3][1][0] = 0.0;
  chrisD[3][3][1][1] = 0.0;
  chrisD[3][3][1][2] = 0.0;
  chrisD[3][3][1][3] = 0.0;
  chrisD[3][3][2][0] = 0.0;
  chrisD[3][3][2][1] = 0.0;
  chrisD[3][3][2][2] = 0.0;
  chrisD[3][3][2][3] = 0.0;
  chrisD[3][3][3][0] = 0.0;
  chrisD[3][3][3][1] = 0.0;
  chrisD[3][3][3][2] = 0.0;
  chrisD[3][3][3][3] = 0.0;

  return true;
}


/*! Transform local 4-direction to coordinate 4-direction.
 *
 *  \param  pos  :  pointer to position array.
 *  \param  ldir :  pointer to local direction array.
 *  \param  dir  :  pointer to calculated coordinate direction array.
 *  \param  type :  type of tetrad.
 */
void
MetricStraightSpinningString :: localToCoord ( const double* pos, const double* ldir, double* dir,
                                               enum_nat_tetrad_type  type  )
{ 
  double rho = pos[1];

  double a = mA;
  double k = mK;
  double c = mSpeedOfLight;

  if (type == enum_nat_tetrad_lnrf)
  {
     double w = sqrt(k*k*rho*rho-a*a);
     dir[0] = w/(k*rho)/c*ldir[0];
     dir[1] = ldir[1];
     dir[2] = -a/(w*k*rho)*ldir[0] + ldir[2]/w;
     dir[3] = ldir[3];
  }
  else
  {
     dir[0] = ldir[0]/c + a/(k*rho*c)*ldir[2];
     dir[1] = ldir[1];
     dir[2] = ldir[2]/(k*rho);
     dir[3] = ldir[3];
  }
  calcClosestApproach(pos,ldir);
}

/*! Transform coordinate 4-direction to local 4-direction.
 *
 *  \param  pos  :  pointer to position array.
 *  \param  cdir :  pointer to coordinate direction.
 *  \param  ldir :  pointer to calculated local direction array.
 *  \param  type :  type of tetrad.
 */
void 
MetricStraightSpinningString :: coordToLocal ( const double* , const double* , double* ,
                                               enum_nat_tetrad_type   )
{
  // TODO
}


/*! Test break condition.
 *
 *  \param pos    : pointer to position array.
 *  \return true  : radial position r < 0.0 or  r^2<=(1.0+eps)*rs^2.
 *  \return false : position is valid.
 */
bool 
MetricStraightSpinningString :: breakCondition ( const double* )
{
  bool br = false;
  return br;
}

#if 0
/*! Calculate right hand side of the geodesic equation in first order form.
 *
 *  \param  y[]   : pointer to position and direction coordinates.
 *  \param  dydx[] : pointer to right side of geodesic equation.
 */
bool
MetricStraightSpinningString :: calcDerivs ( const double y[], double dydx[] )
{
  dydx[0] = y[4];
  dydx[1] = y[5];
  dydx[2] = y[6];
  dydx[3] = y[7];

  /*
  dydx[4] = -2.0*G_t_r_t*y[4]*y[5] - 2.0*G_r_ph_t*y[5]*y[7] - 2.0*G_th_ph_t*y[6]*y[7] - 2.0*G_t_th_t*y[4]*y[6];
  dydx[5] = -G_t_t_r*y[4]*y[4] - 2.0*G_t_ph_r*y[4]*y[7] - G_r_r_r*y[5]*y[5] - G_th_th_r*y[6]*y[6] - G_ph_ph_r*y[7]*y[7];
  dydx[6] = -2.0*G_t_ph_th*y[4]*y[7] - 2.0*G_r_th_th*y[5]*y[6] - G_ph_ph_th*y[7]*y[7];
  dydx[7] = -2.0*G_t_th_ph*y[4]*y[6] - 2.0*G_r_ph_ph*y[5]*y[7] - 2.0*G_th_ph_ph*y[6]*y[7];
  */
  return true;
}
#endif

/*! Set parameter 'pName' to 'val'.
 *
 *  \param pName : parameter name.
 *  \param val : parameter value.
 */
bool 
MetricStraightSpinningString :: setParam ( std::string pName, double val )
{
  Metric::setParam(pName,val);

  if (pName=="a")
    mA = val;
  else if (pName=="k")
    mK = val;
  return true;
}


/*! Effective potential.
 *  \param pos : initial position.
 *  \param cdir : initial four-direction.
 *  \param type : geodesic type.
 *  \param x : abscissa value.
 *  \param val : reference to effective potential value.
 *  \return true : effective potential exists at x.
 */
bool
MetricStraightSpinningString :: effPotentialValue  ( const vec4 pos, const vec4 cdir, enum_geodesic_type type, const double x, double &val )
{
  double kappa = 0.0;
  if (type==enum_geodesic_timelike)
     kappa = -mSign;

  if (x<0.0)
    return false;

  double c = mSpeedOfLight;
  double h1 = c*(c*cdir[0] - mA*cdir[2]);
  double h2 = mA*(c*cdir[0] - mA*cdir[2]) + mK*mK*pos[1]*pos[1]*cdir[2];

  val = pow((h2-mA*h1/c)/(mK*x),2.0) - kappa*c*c;
  return true;
}

/*! Total energy.
 *  \param pos : initial position.
 *  \param cdir : initial four-direction.
 *  \param x : abscissa value.
 *  \param val : reference to total energy value.
 *  \return true : effective potential exists at x.
 */
bool
MetricStraightSpinningString :: totEnergy ( const vec4 , const vec4 cdir, const double , double &val )
{
  double c = mSpeedOfLight;
  double h1 = c*(c*cdir[0] - mA*cdir[2]);

  val = h1*h1/(c*c);
  return true;
}


/*! Generate report.
 */
bool    
MetricStraightSpinningString :: report ( const vec4 pos, const vec4 cdir, std::string &text )
{
  std::stringstream ss;
  ss << "Report for StraightSpinningString metric\n\tcoordinate : (t,rho,phi,z)\n";
  ss << "---------------------------------------------------------\n";
  ss << "  physical units ................................. no\n";

  double c = mSpeedOfLight;
  double h1 = c*(c*cdir[0] - mA*cdir[2]);
  double h2 = mA*(c*cdir[0] - mA*cdir[2]) + mK*mK*pos[1]*pos[1]*cdir[2];
  ss.precision(DEF_FIXED_REPORT_PRECISION);
  ss << "  constant of motion ........................ h1 = " << h1 << std::endl;
  ss << "  constant of motion ........................ h2 = " << h2 << std::endl;

  ss << "\n  point of closest approach ..............rho_ca = " << rho_ca << std::endl;
  text = ss.str();
  return true;
}


// ********************************* protected methods *****************************
/*!
 */
void 
MetricStraightSpinningString :: setStandardValues( )
{
  mInitPos[0] = 0.0; 
  mInitPos[1] = 6.0;
  mInitPos[2] = 0.0;
  mInitPos[3] = 0.0;
  mInitDir[0] = 1.0;
  mInitDir[1] = 0.0;
  mInitDir[2] = 0.0;
   
  mCoordNames[0] = std::string("t");
  mCoordNames[1] = std::string("rho");
  mCoordNames[2] = std::string("phi");
  mCoordNames[3] = std::string("z");
}

void
MetricStraightSpinningString :: calcClosestApproach ( const double* pos, const double* ldir )
{
   double g = ldir[0]/mSpeedOfLight;
   double sk = ldir[2];
   if (g>1.0)
   {
      double beta =sqrt(1.0-1.0/(g*g));
      sk = ldir[2]/(g*beta*mSpeedOfLight);
   }
   rho_ca = pos[1]*sk;
}

} // end namespace m4d
