// -------------------------------------------------------------------------------- 
/*
    geodesic_view.h
 
  Copyright (c) 2009  Thomas Mueller, Frank Grave
 
 
   This file is part of the GeodesicViewer.
 
   The GeodesicViewer is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
 
   The GeodesicViewer is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with the GeodesicViewer.  If not, see <http://www.gnu.org/licenses/>.
*/

/*!  \class  GeodesicView
     \brief  Main GUI of the GeodesicViewer.

*/
// -------------------------------------------------------------------------------- 

#ifndef GEODESIC_VIEW_H
#define GEODESIC_VIEW_H

#include <iostream>
#include <cassert>
#include <QtGui>

#include <QtScript>

#include <gdefs.h>

#include <opengl3d_model.h>
#include <opengl2d_model.h>

#include <doubleedit_util.h>
#include <greek.h>

#include <locted_view.h>
#include <geod_view.h>
#include <draw_view.h>
#include <object_view.h>
#include <prot_view.h>
#include <doc_view.h>
#include <report_view.h>

#include <metric/m4dMetricDatabase.h>
#include <motion/m4dMotionList.h>
#include <extra/m4dObject.h>
#include <utils/utilities.h>
#include <utils/myobject.h>

// ---------------------------------------------------
//    class definition:   GeodesicView
// ---------------------------------------------------
class GeodesicView : public QMainWindow
{
  Q_OBJECT

 public:   
   GeodesicView ( m4d::Object* data, struct_params* par, QMainWindow* parent = 0 );
   ~GeodesicView ( );

// --------- public methods -----------
 public:
  virtual QSize minimumSizeHint ( ) const;
  virtual QSize sizeHint        ( ) const;


// --------- public slots -----------
 public slots:
   void  slot_load_setting  ( );
   void  slot_save_setting  ( );
   void  slot_load_vparams  ( );
   void  slot_save_vparams  ( );
   void  slot_reset_vparams ( );

   void  slot_save_image_2d ( );
   void  slot_save_image_3d ( );

   void  slot_load_all      ( );

   void  slot_prot_dialog   ( );
   void  slot_write_prot    ( );
   void  slot_quit          ( );
   
   void  slot_reset         ( );
   void  slot_doc           ( );
   void  slot_about         ( );

   void  slot_changeDrawActive ( );
  
   void  slot_animate ( );
   
   void  slot_load_objects   ( );
   void  slot_append_objects ( );
   void  slot_save_objects   ( );
   void  slot_show_objects   ( );
   void  slot_objChanged     ( );

   void  slot_show_report    ( );
   void  slot_save_report    ( );

   void  slot_setCurrentMetric   ( );
   void  slot_metricParamChanged ( );
   void  slot_setGeodSolver      ( );   
   
   void  slot_setMaxNumPoints     ( );
   void  slot_setStepsizeControl  ( );
   void  slot_setIntegratorParams ( );
   void  slot_setDrawWithLines    ( );

   void  slot_setConstraintEps    ( );
   void  slot_setResize           ( );

   void  slot_setMaxStepSize      ( );
   void  slot_setMinStepSize      ( );

   void  slot_setSIunits       ( );
   void  slot_setGeomUnits     ( );
   void  slot_setUnits         ( );
   
   void  slot_calcGeodesic     ( );

   void  slot_setOpenGLcolors  ( );

   void  slot_executeScript    ( );

   // script functions
   void  reset          ( );
   void  setMetric      ( QString name );
   void  setMetricParam ( int num, double val );
   void  setMetricParam ( QString name, double val );
   void  setSolver      ( QString name );
   void  setSolverParam ( QString name, double val );
   void  wait           ( double sec );

// --------- protected methods -----------
 protected:  
  void  init           ( );
  void  initElements   ( );
  void  initGUI        ( );
  void  initControl    ( );
  void  initActions    ( );
  void  initMenus      ( );
  void  initScripting  ( );
  void  initStatusTips ( );

  void  calculateGeodesic     ( );
  void  calculateGeodesicData ( );

  bool  setMetric               ( m4d::enum_metric      metric );
  bool  setMetricNamesAndCoords ( );
  bool  setGeodSolver           ( m4d::enum_integrator  type );  

  void  setUnits    ( const double sol, const double grav, const double diperm );

  bool  saveSetting ( QString filename );
  bool  saveObjects ( QString filename );
  bool  saveReport  ( QString filename );

  void  setSetting  ( );
  void  loadObjects ( QString filename );

  void  execScript  ( QString filename );

 protected slots:
   virtual void closeEvent ( QCloseEvent * event );

// -------- private attributes --------
 private:
   struct_params* mParams;

  /* ---- 2D/3D drawing ---- */
  QTabWidget*     tab_draw;
  QWidget*        wgt_opengl;
  OpenGL3dModel*  opengl;
  QWidget*        wgt_2d;

  OpenGL2dModel*  draw2d;
  QAction*        mActionMake3Dactive;
  QAction*        mActionMake2Dactive;

  QLabel*         lab_value;
  QLabel*         lab_step;

  
  /* ---- metric/integrator ---- */
  QTabWidget*     tab_metricInt;
  
  QComboBox*      cob_metric;
  QTableWidget*   tbw_metric_params;
  QWidget*        wgt_metric;
  
  QComboBox*      cob_integrator;
  QCheckBox*      chb_stepsize_controlled;
  QCheckBox*      chb_drawstyle;
  QLabel*         lab_maxpoints;
  QLineEdit*      led_maxpoints;
  QLabel*         lab_stepsize;
  QLineEdit*      led_stepsize;
  QLabel*         lab_max_stepsize;
  QLineEdit*      led_max_stepsize;
  QLabel*         lab_min_stepsize;
  QLineEdit*      led_min_stepsize;
  QLabel*         lab_eps_abs;
  QLineEdit*      led_eps_abs;
  QLabel*         lab_eps_rel;
  QLineEdit*      led_eps_rel;
  QLabel*         lab_constraint_eps;
  QLineEdit*      led_constraint_eps;
  QLabel*         lab_resize_eps;
  QLineEdit*      led_resize_eps;
  QLabel*         lab_resize_fac;
  QLineEdit*      led_resize_fac;
  QWidget*        wgt_integrator;
 
  QLabel*         lab_speed_of_light;
  QLineEdit*      led_speed_of_light;  
  QLabel*         lab_grav_constant;
  QLineEdit*      led_grav_constant;
  QLabel*         lab_dielectric_perm;
  QLineEdit*      led_dielectric_perm;
  QPushButton*    pub_si_units;
  QPushButton*    pub_geom_units;
  QWidget*        wgt_constants;
 
  /* ---- local tetrad ---- */
  LoctedView*     lct_view;

  /* ---- geodesic ---- */  
  GeodView*       geo_view;

  /* ---- draw handing ---- */ 
  DrawView*       drw_view;
  OpenGLJacobiModel*  mOglJacobi;
  
  /* ---- File menu ---- */
  QMenu*          mFileMenu;
  QAction*        mActionReset;
  QAction*        mActionLoadSetting;
  QAction*        mActionSaveSetting;
  QAction*        mActionLoadViewParams;
  QAction*        mActionSaveViewParams;
  QAction*        mActionResetViewParams;
  QAction*        mActionLoadAll;
  QAction*        mActionSaveImage2D;
  QAction*        mActionSaveImage3D;
  QAction*        mActionWriteProtocoll;
  QAction*        mActionQuit;

  /* ---- Object menu ---- */
  QMenu*          mObjectMenu;
  QAction*        mActionLoadObjectFile;
  QAction*        mActionAppendObjectFile;
  QAction*        mActionSaveObjectFile;
  QAction*        mActionShowObjects;

  /* ---- Report menu ---- */
  QMenu*          mReportMenu;
  QAction*        mActionShowReport;
  QAction*        mActionSaveReport;

  /* ---- Help menu ---- */
  QMenu*          mHelpMenu;
  QAction*        mActionAbout;
  QAction*        mActionDoc;
  DocView*        mHelpDoc;

  QAction*        mActionPlay;
  QAction*        mActionStepKsiPlus;
  QAction*        mActionStepKsiMinus;
  QAction*        mActionStepChiPlus;
  QAction*        mActionStepChiMinus;

  ObjectView*     mObjectView;
  ReportDialog*   mReportText;
  ProtDialog*     mProtDialog;
   
  /* ---- QtScript ---- */
  QScriptEngine*  mScriptEngine;
  QAction*        mLoadScript;

  /* ---- status bar ---- */
  QStatusBar*     mStatus;
  QLineEdit*      led_status;
  QLabel*         lab_num_points;
  QLCDNumber*     lcd_num_points;
#ifdef SHOW_CALC_TIME
  QLabel*         lab_calc_time;
  QLCDNumber*     lcd_calc_time;
#endif

  m4d::Object*            mData;
  std::vector<m4d::vec4>  mPointData;
  std::vector<m4d::vec4>  mDirData;
  std::vector<double>     mEpsData;
  std::vector<MyObject*>  mObjects;

  GreekLetter     mGreekLetter;
};

#endif
